# This file is part of the litprog project
# https://github.com/litprog/litprog
#
# Copyright (c) 2018-2021 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
import re
import typing as typ
import logging
import pathlib as pl

logger = logging.getLogger(__name__)


FilePaths = typ.Iterable[pl.Path]

MD_FRONT_MATTER = 'front_matter'

MD_HEADLINE   = 'headline'
MD_PARAGRAPH  = 'paragraph'
MD_LIST       = 'list'
MD_BLOCKQUOTE = 'blockquote'

# https://python-markdown.github.io/extensions/definition_lists/
MD_DEF_LIST = 'def_list'

# https://python-markdown.github.io/extensions/footnotes/
MD_FOOTNOTE_DEF = 'footnote_def'

MD_BLOCK = 'block'

VALID_ELEMENT_TYPES = {
    MD_FRONT_MATTER,
    MD_HEADLINE,
    MD_PARAGRAPH,
    MD_LIST,
    MD_BLOCKQUOTE,
    MD_DEF_LIST,
    MD_FOOTNOTE_DEF,
    MD_BLOCK,
}

NON_CODE_BLOCKS = ('bob', 'math')

MarkdownElementType = str

FrontMatterMetadata = typ.Dict[str, typ.Any]

DEFAULT_FRONT_MATTER_META = {'lang': "en-US", 'title': "-"}

VALID_METADATA_KEYS = {
    'author',
    'copyright',
    'copyright_url',
    'description',
    'favicon_url',
    'keywords',
    'lang',
    'logo_url',
    'project_name',
    'repo_url',
    'title',
}


FRONT_MATTER_PATTERN = r"^(\+\+\+|\-\-\-)$"

FRONT_MATTER_RE = re.compile(FRONT_MATTER_PATTERN, flags=re.MULTILINE)


HEADLINE_PATTERN_A = r"""
    ^
    (?P<headline_marker_a>[#]+)
    (?P<headline_text_a>[^#\n]+)
    ([#]*)?
"""


HEADLINE_PATTERN_B = r"""
    ^
    (?P<headline_text_b>.*)\n
    (?P<headline_marker_b>=+|-+)
"""


BLOCK_START_PATTERN = r"""
    ^
    (?P<block_fence>```|~~~)
    (?P<info_string>[^\n]*)?
"""

ELEMENT_PATTERN = f"""
    (?:{HEADLINE_PATTERN_A})
    | (?:{HEADLINE_PATTERN_B})
    | (?:{BLOCK_START_PATTERN})
"""

# https://regex101.com/r/cSbfvF/35
IMAGE_URL_PATTERN = r"""
!\[
  (?P<alt>[^\]]*)
\]
\(
  (?P<url>.*?)(?=\"|\))
  (?P<title>\".*\")?
\)
"""

LANGUAGE_COMMENT_TEMPLATES = {
    "c++"          : "// {}",
    'actionscript' : "// {}",
    'actionscript3': "// {}",
    'bash'         : "# {}",
    'c'            : "// {}",
    'd'            : "// {}",
    'elixir'       : "# {}",
    'erlang'       : "% {}",
    'go'           : "// {}",
    'zig'          : "// {}",
    'java'         : "// {}",
    'javascript'   : "// {}",
    'json'         : "// {}",
    'swift'        : "// {}",
    'r'            : "// {}",
    'php'          : "// {}",
    'svg'          : "<!-- {} -->",
    'html'         : "<!-- {} -->",
    'css'          : "/* {} */",
    'csharp'       : "// {}",
    'fsharp'       : "// {}",
    'kotlin'       : "// {}",
    'make'         : "# {}",
    'nim'          : "# {}",
    'perl'         : "# {}",
    'yaml'         : "# {}",
    'prolog'       : "% {}",
    'scheme'       : "; {}",
    'clojure'      : "; {}",
    'lisp'         : "; {}",
    'coffee-script': "# {}",
    'python'       : "# {}",
    'ruby'         : "# {}",
    'rust'         : "// {}",
    'scala'        : "// {}",
    'sh'           : "# {}",
    'shell'        : "# {}",
    'sql'          : "-- {}",
    'typescript'   : "// {}",
}


LANGUAGE_COMMENT_PATTERNS = {
    "c++"          : (r"^[ \t]*//"  , r"$"),
    'actionscript' : (r"^[ \t]*//"  , r"$"),
    'actionscript3': (r"^[ \t]*//"  , r"$"),
    'bash'         : (r"^[ \t]*[#]" , r"$"),
    'c'            : (r"^[ \t]*//"  , r"$"),
    'd'            : (r"^[ \t]*//"  , r"$"),
    'elixir'       : (r"^[ \t]*[#]" , r"$"),
    'erlang'       : (r"^[ \t]*%"   , r"$"),
    'go'           : (r"^[ \t]*//"  , r"$"),
    'zig'          : (r"^[ \t]*//"  , r"$"),
    'java'         : (r"^[ \t]*//"  , r"$"),
    'javascript'   : (r"^[ \t]*//"  , r"$"),
    'json'         : (r"^[ \t]*//"  , r"$"),
    'swift'        : (r"^[ \t]*//"  , r"$"),
    'r'            : (r"^[ \t]*//"  , r"$"),
    'php'          : (r"^[ \t]*//"  , r"$"),
    'svg'          : (r"^[ \t]*<!--", r"-->$"),
    'html'         : (r"^[ \t]*<!--", r"-->$"),
    'css'          : (r"^[ \t]*/\*" , r"\*/$"),
    'csharp'       : (r"^[ \t]*//"  , r"$"),
    'fsharp'       : (r"^[ \t]*//"  , r"$"),
    'kotlin'       : (r"^[ \t]*//"  , r"$"),
    'make'         : (r"^[ \t]*[#]" , r"$"),
    'nim'          : (r"^[ \t]*[#]" , r"$"),
    'perl'         : (r"^[ \t]*[#]" , r"$"),
    'yaml'         : (r"^[ \t]*[#]" , r"$"),
    'prolog'       : (r"^[ \t]*%"   , r"$"),
    'scheme'       : (r"^[ \t]*;"   , r"$"),
    'clojure'      : (r"^[ \t]*;"   , r"$"),
    'lisp'         : (r"^[ \t]*;"   , r"$"),
    'coffee-script': (r"^[ \t]*[#]" , r"$"),
    'python'       : (r"^[ \t]*[#]" , r"$"),
    'ruby'         : (r"^[ \t]*[#]" , r"$"),
    'rust'         : (r"^[ \t]*//"  , r"$"),
    'scala'        : (r"^[ \t]*//"  , r"$"),
    'sh'           : (r"^[ \t]*[#]" , r"$"),
    'shell'        : (r"^[ \t]*[#]" , r"$"),
    'sql'          : (r"^[ \t]*--"  , r"$"),
    'typescript'   : (r"^[ \t]*//"  , r"$"),
}


def _re(pattern: str, flags: int = re.MULTILINE) -> typ.Pattern:
    return re.compile(pattern, flags=flags)


LANGUAGE_COMMENT_REGEXES = {
    lang: (_re(start_pattern), _re(end_pattern))
    for lang, (start_pattern, end_pattern) in LANGUAGE_COMMENT_PATTERNS.items()
}

KNOWN_INFO_STRINGS = {
    'math',
    'bob',
    'aafigure',
}

KNOWN_INFO_STRINGS.update(LANGUAGE_COMMENT_REGEXES.keys())


ELEMENT_RE = _re(ELEMENT_PATTERN, flags=re.VERBOSE | re.MULTILINE)

HEADLINE_RE_A = _re(HEADLINE_PATTERN_A, flags=re.VERBOSE | re.MULTILINE)
HEADLINE_RE_B = _re(HEADLINE_PATTERN_B, flags=re.VERBOSE | re.MULTILINE)

BLOCK_START_RE = _re(BLOCK_START_PATTERN, flags=re.VERBOSE | re.MULTILINE)

BLOCK_END_RE = {
    "```": _re(r"^```", flags=re.VERBOSE | re.MULTILINE),
    "~~~": _re(r"^~~~", flags=re.VERBOSE | re.MULTILINE),
}

IMAGE_URL_RE = _re(IMAGE_URL_PATTERN, flags=re.VERBOSE | re.MULTILINE)


class ImageTag(typ.NamedTuple):
    alt  : str
    url  : str
    title: str


class _RawMarkdownElement(typ.NamedTuple):

    md_type   : MarkdownElementType
    first_line: int
    content   : str


# NOTE (mb 2019-05-30): The word "Successor" refers
#   to the relationship of a MarkdownElement to a
#   modified version of itself. It does _not_ refer to
#   it's document position relative to another.
Successor = typ.Optional['MarkdownElement']


class MarkdownElement:

    md_path   : pl.Path
    first_line: int
    elem_index: int
    md_type   : MarkdownElementType
    content   : str
    _successor: typ.Optional[typ.Any]

    # Recursive types not fully supported yet;
    # this class can be changed to a NamedTuple once they are.
    # Successor  : typ.Optional['MarkdownElement']
    @property
    def successor(self) -> Successor:
        return typ.cast(Successor, self._successor)

    def __init__(
        self,
        md_path   : pl.Path,
        first_line: int,
        elem_index: int,
        md_type   : MarkdownElementType,
        content   : str,
        successor : Successor,
    ) -> None:
        assert md_type in VALID_ELEMENT_TYPES
        self.md_path    = md_path
        self.first_line = first_line
        self.elem_index = elem_index
        self.md_type    = md_type
        self.content    = content
        self._successor = successor

    def __repr__(self) -> str:
        addr = hex(id(self))
        return (
            "MarkdownElement("
            + f"'{self.md_path}', "
            + f"{self.first_line}, "
            + f"{self.elem_index}, "
            + f"'{self.md_type}', "
            + f"...) at {addr}"
        )

    def clone(self) -> 'MarkdownElement':
        return MarkdownElement(
            md_path=self.md_path,
            first_line=self.first_line,
            elem_index=self.elem_index,
            md_type=self.md_type,
            content=self.content,
            successor=self._successor or self,
        )


MarkdownElements      = typ.List[MarkdownElement]
MaybeMarkdownElements = typ.Optional[MarkdownElements]


class Headline(typ.NamedTuple):

    md_path   : pl.Path
    elem_index: int
    text      : str
    level     : int


InfoString = str


class Directive(typ.NamedTuple):

    name : str
    value: str

    raw_text: str


class Block(typ.NamedTuple):

    md_path           : pl.Path
    namespace         : str
    first_line        : int
    elem_index        : int
    info_string       : InfoString
    directives        : typ.List[Directive]
    content           : str
    inner_content     : str
    includable_content: str


# NOTE (mb 2021-03-04): Right now the 'out' directive references the
#   preceeding block. We could also make it more general by allowing
#   explicit references to the def of an exec block. This would
#   be similar to requires.

VALID_DIRECTIVES = {
    'language',
    # block composition
    'def',
    'addto',
    'dep',
    'include',
    # session/subprocess
    'exec',
    'out',
    'run',
    # parameters for out and run
    'debug',
    'expect',
    'timeout',
    # NOTE: input_delay might allow the accurate
    #   association of input/output as long as output
    #   is always captured by the time the delay passes.
    'input_delay',
    'proc_info',
    'out_prefix',
    'err_prefix',
    'out_color',
    'err_color',
    # file generation
    'file',
    # build system
    'requires',  # comma separated globs for ids to invalidate block
    # 'cache'   # yes|once|never
    # 'stateful',
    # 'pure',
    # 'make',
    # 'const'
    # 'use_macro',
    # 'def_macro',
}


VALID_NOARG_DIRECTIVES  = {'exec', 'out', 'debug'}
VALID_ARG_DIRECTIVES    = VALID_DIRECTIVES - {'out', 'debug'}
VALID_INLINE_DIRECTIVES = {'dep', 'include'}


assert VALID_INLINE_DIRECTIVES < VALID_ARG_DIRECTIVES
assert VALID_NOARG_DIRECTIVES  < VALID_DIRECTIVES


def get_line_directive(line: str, language: typ.Optional[str], is_prelude: bool) -> typ.Optional[str]:
    if language is None:
        comment_text = line
    else:
        comment_start_re, comment_end_re = LANGUAGE_COMMENT_REGEXES[language]
        start_match = comment_start_re.search(line)
        if start_match is None:
            return None

        comment_text = line[start_match.end() :]
        end_match    = comment_end_re.search(comment_text)
        if end_match is None:
            return None

    comment_text = comment_text.strip()
    if is_prelude:
        if comment_text in VALID_NOARG_DIRECTIVES:
            return comment_text

        for name in VALID_ARG_DIRECTIVES:
            if comment_text.startswith(name + ":"):
                return name
    else:
        for name in VALID_INLINE_DIRECTIVES:
            if comment_text.startswith(name + ":"):
                return name

    return None


def _parse_directive(directive_text: str, raw_text: str) -> Directive:
    if ":" in directive_text:
        name, value = directive_text.split(":", 1)
        name  = name.strip()
        value = value.strip()
    else:
        name  = directive_text.strip()
        value = ""

    if name in VALID_DIRECTIVES:
        return Directive(name, value, raw_text)
    else:
        errmsg = f"Invalid directive '{name}'"
        raise Exception(errmsg)


# NOTE (mb 2020-05-22): Since we do multiple passes over the file, we
#   avoid reporting

FileLocation = str


class ParseError(typ.NamedTuple):

    location: FileLocation
    level   : int
    message : str


AnyElem = typ.Union[MarkdownElement, Block]


def location(elem: AnyElem) -> FileLocation:
    return f"Line {elem.first_line:<4} of {elem.md_path}"


def make_parse_error(message: str, elem: AnyElem, level: int = logging.ERROR) -> ParseError:
    return ParseError(location(elem), level, message)


FILENAME_PATTERN_URL = "https://regex101.com/r/sLzB5p/4"

# Input files for LitProg must match this pattern.
# The 'namespace' group is used to reference blocks
# in different files in a way that doesn't break if
# files are reordered.

FILENAME_PATTERN = r"""
^
(?:[0-9_\-\s]+)?
(?P<namespace>[\w ]+)
(?:\.\w*)?
$
"""

FILENAME_RE = re.compile(FILENAME_PATTERN, flags=re.VERBOSE)


class MarkdownFile:

    md_path : pl.Path
    errors  : typ.Set[ParseError]
    elements: MarkdownElements

    def __init__(self, md_path: pl.Path, elements: MaybeMarkdownElements = None) -> None:
        self.md_path = md_path
        self.errors  = set()
        if elements is None:
            self.elements = _parse_md_elements(md_path)
        else:
            self.elements = elements

    def copy(self) -> 'MarkdownFile':
        return MarkdownFile(self.md_path, list(self.elements))

    def block_namespace(self) -> str:
        # The namespace is based only on the filename, directories are
        # only for organization, otherwise they are ignored.
        filename       = self.md_path.name
        filename_match = FILENAME_RE.match(filename)
        if filename_match is None:
            errmsg = f"Invalid filename {filename}, must match {FILENAME_PATTERN_URL}"
            raise Exception(errmsg)
        else:
            raw_namespace        = filename_match.group('namespace')
            normalized_namespace = raw_namespace.replace("-", "_").replace(" ", "_")
            return normalized_namespace

    def headlines(self) -> typ.Iterable[Headline]:
        for elem_index, elem in enumerate(self.elements):
            if elem.md_type != MD_HEADLINE:
                continue

            a_match = HEADLINE_RE_A.match(elem.content)
            b_match = HEADLINE_RE_B.match(elem.content)
            if a_match:
                text   = a_match.group('headline_text_a')
                marker = a_match.group('headline_marker_a')
                level  = marker.count("#")
            elif b_match:
                text   = b_match.group('headline_text_b')
                marker = b_match.group('headline_marker_b')
                level  = 1 if "-" in marker else 2
            else:
                err_msg = "Invalid headline: {elem.content}"
                raise ValueError(err_msg)

            yield Headline(self.md_path, elem_index, text.strip(), level)

    def image_tags(self) -> typ.Iterable[ImageTag]:
        for elem in self.elements:
            if elem.md_type in (MD_HEADLINE, MD_BLOCK):
                continue
            for image_match in IMAGE_URL_RE.finditer(elem.content):
                yield ImageTag(*image_match.groups())

    def _init_plain_block(self, elem: MarkdownElement, info_string: str) -> Block:
        inner_content = elem.content
        inner_content = inner_content.split("\n", 1)[-1]
        # trim off final fence
        inner_content = inner_content.rsplit("\n", 1)[0]

        return Block(
            self.md_path,
            self.block_namespace(),
            elem.first_line,
            elem.elem_index,
            info_string,
            [],
            elem.content,
            # TODO (mb 2020-06-02): Why do we .strip() here ?
            inner_content.strip(),
            inner_content.strip(),
        )

    def _init_code_block(self, elem: MarkdownElement, info_string: str, rest_content: str) -> Block:
        language = info_string
        comment_start_re, comment_end_re = LANGUAGE_COMMENT_REGEXES[language]

        directives = [Directive('language', language, language)]

        inner_chunks      = []
        includable_chunks = []

        is_prelude = True
        rest       = rest_content
        while rest:
            start_match = comment_start_re.search(rest)

            if start_match is None:
                inner_chunks.append(rest)
                includable_chunks.append(rest)
                break

            prefix_chunk = rest[: start_match.start()]
            if prefix_chunk:
                inner_chunks.append(prefix_chunk)
                includable_chunks.append(prefix_chunk)

            rest = rest[start_match.end() :]

            if prefix_chunk.strip() and is_prelude:
                # prelude ends if there is non-whitespace before a comment
                is_prelude = False

            end_match = comment_end_re.search(rest)
            if end_match is None:
                comment_text = rest
                rest         = ""
            else:
                comment_text = rest[: end_match.start()]
                rest         = rest[end_match.end() :]

            raw_text = start_match.group(0) + comment_text
            assert raw_text in elem.content

            comment_text = comment_text.strip()

            directive_name = get_line_directive(comment_text, language=None, is_prelude=is_prelude)
            if directive_name:
                directive = _parse_directive(comment_text, raw_text)
                directives.append(directive)
                inner_chunks.append(raw_text)
                if directive.name in ('dep', 'include'):
                    # NOTE (mb 2020-06-03): needed for recursive include
                    includable_chunks.append(raw_text)
            else:
                inner_chunks.append(raw_text)
                includable_chunks.append(raw_text)

        inner_content = "".join(inner_chunks)
        # trim off final fence
        inner_content = inner_content.rsplit("\n", 1)[0]
        inner_content = "\n".join(line for line in inner_content.splitlines() if line.strip())

        includable_content = "".join(includable_chunks)
        # trim off final fence
        includable_content = includable_content.rsplit("\n", 1)[0]
        includable_content = "\n".join(line for line in includable_content.splitlines() if line.strip())

        return Block(
            self.md_path,
            self.block_namespace(),
            elem.first_line,
            elem.elem_index,
            info_string,
            directives,
            elem.content,
            inner_content,
            includable_content,
        )

    def _iter_block_elements(self) -> typ.Iterable[typ.Tuple[MarkdownElement, str, str]]:
        for elem in self.elements:
            if elem.md_type == 'block':
                start_match = BLOCK_START_RE.match(elem.content)
                assert start_match is not None
                maybe_info_string = start_match.group('info_string')
                info_string       = typ.cast(str, maybe_info_string or "")
                info_string       = info_string.strip()
                content_start     = start_match.end()
                content           = elem.content[content_start:]
                yield (elem, info_string, content)

    def iter_blocks(self) -> typ.Iterable[Block]:
        for elem, info_string, content in self._iter_block_elements():
            is_known_info_string = info_string in KNOWN_INFO_STRINGS
            if info_string.strip() and not is_known_info_string:
                err = make_parse_error(f"Unknown language '{info_string}'", elem, logging.WARNING)
                self.errors.add(err)

            is_valid_language = info_string in LANGUAGE_COMMENT_REGEXES
            if is_valid_language:
                yield self._init_code_block(elem, info_string, rest_content=content)
            else:
                yield self._init_plain_block(elem, info_string)

    def iter_block_linenos(self) -> typ.Iterable[typ.Tuple[int, int]]:
        for elem, info_string, content in self._iter_block_elements():
            if info_string not in NON_CODE_BLOCKS:
                num_lines = content.strip().count("\n")
                yield elem.first_line, num_lines

    def parse_front_matter_meta(self) -> FrontMatterMetadata:
        # pylint: disable=import-outside-toplevel;
        #   to speed up cli, lazy load modules: yaml and toml

        metadata = DEFAULT_FRONT_MATTER_META.copy()
        if len(self.elements) > 0 and self.elements[0].md_type == MD_FRONT_MATTER:
            content = self.elements[0].content
            if content.startswith("+++"):
                import toml

                meta = toml.loads(content.strip("+ \n\r"))
            elif content.startswith("---"):
                import yaml

                meta = yaml.safe_load(content.strip("- \n\r"))
            else:
                raise RuntimeError("Invalid front matter")

            for key, val in metadata.items():
                if key not in VALID_METADATA_KEYS:
                    errmsg = f"Invalid key Markdown front matter: {key}={val}"
                    raise KeyError(errmsg)

            metadata.update(meta)
        return metadata

    def __lt__(self, other: 'MarkdownFile') -> bool:
        return self.md_path < other.md_path

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, MarkdownFile):
            return False
        elif not self.md_path == other.md_path:
            return False
        elif not self.elements == other.elements:
            return False
        else:
            return True

    def __str__(self) -> str:
        return "".join(elem.content for elem in self.elements if elem.md_type != MD_FRONT_MATTER)

    def __repr__(self) -> str:
        return f"litprog.parse.MarkdownFile(\"{self.md_path}\")"

    def __hash__(self) -> int:
        return hash(self.md_path)


def _iter_raw_md_elements(content: str) -> typ.Iterable[_RawMarkdownElement]:
    line_no = 1
    while content:
        if line_no == 1:
            start_match = FRONT_MATTER_RE.match(content)
            if start_match and start_match.start() == 0:
                end_match = FRONT_MATTER_RE.search(content, pos=start_match.end())
                if end_match:
                    front_matter_end     = end_match.end()
                    front_matter_content = content[:front_matter_end]
                    yield _RawMarkdownElement(MD_FRONT_MATTER, line_no, front_matter_content)
                    line_no += front_matter_content.count("\n")
                    content = content[front_matter_end:]

        match = ELEMENT_RE.search(content)
        if match is None:
            break

        # yield preceding paragraph
        para_content = content[: match.start()]
        if para_content:
            yield _RawMarkdownElement(MD_PARAGRAPH, line_no, para_content)

        line_no += para_content.count("\n")

        # parse match as special element
        groups         = match.groupdict()
        is_headline    = bool(groups['headline_marker_a'] or groups['headline_marker_b'])
        is_block_fence = groups['block_fence']
        match_content  = content[match.start() : match.end()]
        rest_content   = content[match.end() :]

        if is_headline:
            md_type = MD_HEADLINE
        elif is_block_fence:
            md_type      = MD_BLOCK
            block_fence  = groups['block_fence']
            block_end_re = BLOCK_END_RE[block_fence]
            end_match    = block_end_re.search(rest_content)
            if end_match is None:
                match_content += rest_content
                rest_content = ""
            else:
                end_pos = end_match.end()
                match_content += rest_content[:end_pos]
                rest_content = rest_content[end_pos:]

        yield _RawMarkdownElement(md_type, line_no, match_content)

        line_no += match_content.count("\n")
        content = rest_content

    if content:
        yield _RawMarkdownElement(MD_PARAGRAPH, line_no, content)


def _parse_md_elements(md_path: pl.Path) -> MarkdownElements:
    # TODO: encoding from config
    with md_path.open(mode='r', encoding="utf-8") as fobj:
        content = fobj.read()

    elements = []
    for elem_index, raw_elem in enumerate(_iter_raw_md_elements(content)):
        elem = MarkdownElement(
            md_path,
            raw_elem.first_line,
            elem_index,
            raw_elem.md_type,
            raw_elem.content,
            None,
        )
        elements.append(elem)

    # An important criteria for the context is that it has the
    # complete text of the original literate program and is able to
    # reproduce it byte for byte. This must be possible, because we
    # want to be able to update elements of the literate program and
    # write it back to disk.

    assert content == "".join(elem.content for elem in elements)
    return elements


MarkdownFiles = typ.List[MarkdownFile]


class Context:

    files: MarkdownFiles

    def __init__(self, md_path_or_files: typ.Union[FilePaths, MarkdownFiles]) -> None:
        self.files = []
        for path_or_file in md_path_or_files:
            if isinstance(path_or_file, MarkdownFile):
                self.files.append(path_or_file)
            else:
                self.files.append(MarkdownFile(path_or_file))
        self.files.sort()

    def headlines(self) -> typ.Iterable[Headline]:
        for md_file in self.files:
            for headline in md_file.headlines():
                yield headline

    def iter_blocks(self) -> typ.Iterable[Block]:
        for md_file in self.files:
            for block in md_file.iter_blocks():
                yield block

    def copy(self) -> 'Context':
        return Context([f.copy() for f in self.files])

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Context) and self.files == other.files


def parse_context(md_paths: FilePaths) -> Context:
    parse_ctx = Context(md_paths)

    # provoke parse errors early on
    assert parse_ctx.copy() == parse_ctx
    list(parse_ctx.headlines())
    list(parse_ctx.iter_blocks())

    for md_file in parse_ctx.files:
        for err in md_file.errors:
            logger.log(err.level, f"{err.location:<3} : " + err.message)

    return parse_ctx



# This file is part of the litprog project
# https://github.com/litprog/litprog
#
# Copyright (c) 2018-2021 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
import sys
import logging
import pathlib as pl


def html2pdf(html_text: str, out_path: pl.Path, html_dir: pl.Path) -> None:
    # pylint: disable=import-outside-toplevel ; lazy import since we don't always need it
    #   the lazy import makes for a better cli experience
    import weasyprint

    logging.getLogger('weasyprint').setLevel(logging.ERROR)

    wp_ctx = weasyprint.HTML(string=html_text, base_url=str(html_dir))
    with out_path.open(mode="wb") as fobj:
        wp_ctx.write_pdf(fobj)


def main(in_path: pl.Path, out_path: pl.Path) -> None:
    with in_path.open(mode="rt") as in_fobj:
        html2pdf(in_fobj.read(), out_path, in_path.parent)


if __name__ == '__main__':
    main(pl.Path(sys.argv[1]), pl.Path(sys.argv[2]))



# This file is part of the litprog project
# https://github.com/litprog/litprog
#
# Copyright (c) 2018-2021 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
import os
import re
import sys
import time
import typing as typ
import logging
import pathlib as pl
import tempfile
import collections
from concurrent.futures import Future
from concurrent.futures import ThreadPoolExecutor

from . import cache
from . import parse
from . import common
from . import session

logger = logging.getLogger(__name__)


COLOR_CODE_RE = re.compile(r"\d+(;\d)?")


TERM_COLORS = {
    'black'  : "30",
    'red'    : "31",
    'green'  : "32",
    'yellow' : "33",
    'blue'   : "34",
    'magenta': "35",
    'cyan'   : "36",
    'white'  : "37",
}

DEFUALT_TIMEOUT = 9.0

MarkdownFiles = typ.Iterable[parse.MarkdownFile]

# The expanded files are structurally no different, it's just that
# their elements are expanded.
ExpandedMarkdownFile  = parse.MarkdownFile
ExpandedMarkdownFiles = typ.Iterable[ExpandedMarkdownFile]


class BlockError(Exception):

    block           : parse.Block
    include_contents: typ.List[str]

    def __init__(
        self, msg: str, block: parse.Block, include_contents: typ.Optional[typ.List[str]] = None
    ) -> None:
        self.block            = block
        self.include_contents = include_contents or []

        loc = parse.location(block).strip()
        super().__init__(f"{loc} - " + msg)


def get_directive(
    block: parse.Block, name: str, missing_ok: bool = True, many_ok: bool = True
) -> typ.Optional[parse.Directive]:
    found: typ.List[parse.Directive] = []
    for directive in block.directives:
        if directive.name == name:
            if many_ok:
                return directive
            else:
                found.append(directive)

    if len(found) == 0:
        if missing_ok:
            return None
        else:
            errmsg = f"Could not find block with expected '{name}'"
            raise BlockError(errmsg, block)
    elif len(found) > 1 and not many_ok:
        errmsg = f"Block with multiple '{name}'"
        raise BlockError(errmsg, block)
    else:
        return found[0]


def iter_directives(block: parse.Block, name: str) -> typ.Iterable[parse.Directive]:
    for directive in block.directives:
        if directive.name == name:
            yield directive


def _match_indent(directive_text: str, include_val: str) -> str:
    unindented = directive_text.lstrip()
    indent     = directive_text[: -len(unindented)].strip("\n")
    if indent:
        return "\n".join(indent + line for line in include_val.splitlines())
    else:
        return include_val


def _indented_include(
    content       : str,
    directive_text: str,
    include_val   : str,
    is_dep        : bool = True,
) -> str:
    if is_dep:
        # dependencies are only included once (at the first occurance of a dep directive)
        return content.replace(directive_text, include_val, 1).replace(directive_text, "")
    else:
        return content.replace(directive_text, include_val)


# TODO: this should be part of the parsing of all directives
def _parse_prefix(directive: parse.Directive) -> str:
    val = directive.value.strip()
    if val.startswith("'") and val.endswith("'"):
        val = val[1:-1]
    if val.startswith('"') and val.endswith('"'):
        val = val[1:-1]
    return val


BlockId       = str
ScopedBlockId = str

BlockIds       = typ.List[BlockId]
BlockListBySid = typ.Dict[ScopedBlockId, typ.List[parse.Block  ]]
DependencyMap  = typ.Dict[ScopedBlockId, typ.List[ScopedBlockId]]


def _namespaced_lp_id(block: parse.Block, lp_id: BlockId) -> ScopedBlockId:
    if "." in lp_id:
        return lp_id.strip()
    else:
        return block.namespace + "." + lp_id.strip()


def _iter_dep_sids(block: parse.Block) -> typ.Iterable[ScopedBlockId]:
    for lp_dep in iter_directives(block, 'dep'):
        for dep_id in lp_dep.value.split(","):
            yield _namespaced_lp_id(block, dep_id)


def _build_dep_map(blocks_by_sid: BlockListBySid) -> DependencyMap:
    """Build a mapping of block_ids to the Blocks they depend on.

    The mapped block_ids (keys) are only the direct (non-recursive) dependencies.
    """
    dep_map: DependencyMap = {}
    for lp_def_id, blocks in blocks_by_sid.items():
        for block in blocks:
            for dep_sid in _iter_dep_sids(block):
                if dep_sid in blocks_by_sid:
                    if lp_def_id in dep_map:
                        dep_map[lp_def_id].append(dep_sid)
                    else:
                        dep_map[lp_def_id] = [dep_sid]
                else:
                    raise BlockError(f"Unknown block id: {dep_sid}", block)

    if logger.isEnabledFor(logging.DEBUG):
        logger.debug("include map")
        for lp_id, dep_ids in sorted(dep_map.items()):
            dep_ids_str = ", ".join(sorted(dep_ids))
            logger.debug(f" deps  {lp_id:<20}: {dep_ids_str}")

    return dep_map


def _get_dep_cycle(
    lp_id  : BlockId,
    dep_map: DependencyMap,
    root_id: BlockId,
    depth  : int = 0,
) -> typ.List[str]:
    if lp_id in dep_map:
        dep_sids = dep_map[lp_id]
        if root_id in dep_sids:
            return [lp_id]

        for _dep_sid in dep_sids:
            cycle_ids = _get_dep_cycle(_dep_sid, dep_map, root_id, depth + 1)
            if cycle_ids:
                return [lp_id] + cycle_ids

    return []


def _err_on_include_cycle(
    block        : parse.Block,
    lp_id        : BlockId,
    blocks_by_sid: BlockListBySid,
    dep_map      : DependencyMap,
    root_id      : BlockId,
) -> None:
    cycle_ids = _get_dep_cycle(lp_id, dep_map, root_id=root_id)
    if not cycle_ids:
        return

    cycle_ids = [root_id] + cycle_ids
    for cycle_id in cycle_ids:
        cycle_block = blocks_by_sid[cycle_id][0]
        loc         = parse.location(cycle_block).strip()
        logger.warning(f"{loc} - {cycle_id} (trace for include cycle)")

    path   = " -> ".join(f"'{cycle_id}'" for cycle_id in cycle_ids)
    errmsg = f"dep/include cycle {path}"
    raise BlockError(errmsg, block)


def _expand_block_content(
    blocks_by_sid: BlockListBySid,
    block        : parse.Block,
    added_deps   : typ.Set[str],
    dep_map      : DependencyMap,
    keep_fence   : bool,
    lvl          : int = 1,
) -> str:
    # NOTE (mb 2020-12-20): Depth first expansion of content.
    #   This ensures that the first occurance of an dep
    #   directive is expanded at the earliest possible point
    #   even if it is a recursive dep.
    lp_def = get_directive(block, 'def', missing_ok=True, many_ok=False)

    if keep_fence:
        new_content = block.content
    else:
        new_content = block.includable_content + "\n"

    for directive in block.directives:
        if directive.name not in ('dep', 'include'):
            continue

        is_dep = directive.name == 'dep'

        lp_dep_contents: typ.List[str] = []
        for lp_dep_id in directive.value.split(","):
            lp_dep_sid = _namespaced_lp_id(block, lp_dep_id)
            if lp_def:
                lp_def_id = _namespaced_lp_id(block, lp_def.value)
                _err_on_include_cycle(block, lp_dep_sid, blocks_by_sid, dep_map, root_id=lp_def_id)

            if is_dep:
                if lp_dep_sid in added_deps:
                    # skip already included dependencies
                    continue
                else:
                    added_deps.add(lp_dep_sid)

            for dep_block in blocks_by_sid[lp_dep_sid]:
                dep_content = _expand_block_content(
                    blocks_by_sid, dep_block, added_deps, dep_map, keep_fence=False, lvl=lvl + 1
                )
                dep_content = _match_indent(directive.raw_text, dep_content)
                lp_dep_contents.append(dep_content)

        include_content = "\n" + "\n".join(lp_dep_contents)
        dep_text        = directive.raw_text.lstrip("\n")
        new_content     = _indented_include(new_content, dep_text, include_content, is_dep=is_dep)

    return new_content


def _expand_directives(blocks_by_sid: BlockListBySid, md_file: parse.MarkdownFile) -> parse.MarkdownFile:
    new_md_file = md_file.copy()
    dep_map     = _build_dep_map(blocks_by_sid)

    for block in list(new_md_file.iter_blocks()):
        added_deps: typ.Set[str] = set()
        new_content = _expand_block_content(blocks_by_sid, block, added_deps, dep_map, keep_fence=True)
        if new_content != block.content:
            elem = new_md_file.elements[block.elem_index]
            new_md_file.elements[block.elem_index] = parse.MarkdownElement(
                elem.md_path, elem.first_line, elem.elem_index, elem.md_type, new_content, None
            )

    return new_md_file


def _get_blocks_by_id(md_files: MarkdownFiles) -> BlockListBySid:
    blocks_by_sid: BlockListBySid = {}

    for md_file in md_files:
        for block in md_file.iter_blocks():
            lp_def = get_directive(block, 'def', missing_ok=True, many_ok=False)
            if lp_def:
                lp_def_id = lp_def.value
                if "." in lp_def_id and not lp_def_id.startswith(block.namespace + "."):
                    errmsg = f"Invalid block id: {lp_def_id} for namespace {block.namespace}"
                    raise BlockError(errmsg, block)

                block_sid = _namespaced_lp_id(block, lp_def_id)
                if block_sid in blocks_by_sid:
                    prev_block = blocks_by_sid[block_sid][0]
                    prev_loc   = parse.location(prev_block).strip()
                    errmsg     = f"Block already defined: {lp_def_id} at {prev_loc}"
                    raise BlockError(errmsg, block)

                blocks_by_sid[block_sid] = [block]

            for lp_addto in iter_directives(block, 'addto'):
                lp_addto_id = _namespaced_lp_id(block, lp_addto.value)
                if lp_addto_id in blocks_by_sid:
                    blocks_by_sid[lp_addto_id].append(block)
                else:
                    errmsg = f"Unknown block id: {lp_addto_id}"
                    raise BlockError(errmsg, block)

    return blocks_by_sid


def _iter_expanded_files(md_files: MarkdownFiles) -> ExpandedMarkdownFiles:
    # NOTE (mb 2020-05-24): To do the expansion, we have to first
    #   build a graph so that we can resolve blocks for each dep/include.

    # pass 1. collect all blocks (globally) with def directives
    # NOTE (mb 2020-05-31): block ids are always absulute/fully qualified
    blocks_by_sid = _get_blocks_by_id(md_files)

    # pass 2. expand dep directives in markdown files
    for md_file in md_files:
        yield _expand_directives(blocks_by_sid, md_file)


def _iter_block_errors(orig_ctx: parse.Context, build_ctx: parse.Context) -> typ.Iterable[str]:
    """Validate that expansion worked correctly."""
    # NOTE: the main purpose of the orig_ctx is to produce
    #   better error messages. It allows us to point at the
    #   original block, whereas the build_ctx has been
    #   modified and line numbers no longer correspond to
    #   the original file.
    assert len(orig_ctx.files) == len(build_ctx.files)
    for orig_md_file, md_file in zip(orig_ctx.files, build_ctx.files):
        assert orig_md_file.md_path == md_file.md_path
        assert len(orig_md_file.elements) == len(md_file.elements)

        for block in md_file.iter_blocks():
            orig_elem = orig_md_file.elements[block.elem_index]
            for directive in block.directives:
                if directive.name not in ('dep', 'include'):
                    continue

                elem = md_file.elements[block.elem_index]

                rel_line_no = 0
                for line in orig_elem.content.splitlines():
                    if directive.raw_text in line:
                        break

                    rel_line_no += 1

                # TODO (mb 2020-12-30): These line numbers appear to be wrong,
                #   I think for recursive dep directives in particular.
                line_no       = elem.first_line + rel_line_no
                raw_text_repr = repr(directive.raw_text.strip("\n"))
                yield (
                    f"Error processing {md_file.md_path} on line {line_no}: "
                    + f"Could not expand {raw_text_repr}"
                )


def _dump_files(build_ctx: parse.Context) -> None:
    for md_file in build_ctx.files:
        for block in md_file.iter_blocks():
            file_directive = get_directive(block, 'file')
            if file_directive is None:
                continue

            path             = pl.Path(file_directive.value)
            new_content_data = block.includable_content.encode("utf-8")

            if path.exists():
                with path.open(mode="rb") as fobj:
                    old_content_data = fobj.read()

                if old_content_data == new_content_data:
                    # don't needlessly update mtimes
                    continue

            path.parent.mkdir(parents=True, exist_ok=True)
            with path.open(mode="wb") as fobj:
                fobj.write(new_content_data)


def _parse_out_fmt(block: parse.Block, directive_name: str, default_fmt: str) -> str:
    color_directive = get_directive(block, directive_name)
    if color_directive is None:
        return default_fmt

    color_val = color_directive.value.strip()
    if color_val == 'none':
        return default_fmt
    elif color_val in TERM_COLORS:
        color_code = TERM_COLORS[color_val]
        return "\u001b[" + color_code + "m{0}\u001b[0m"
    elif COLOR_CODE_RE.match(color_val):
        return "\u001b[" + color_val + "m{0}\u001b[0m"
    else:
        valid_codes = ", ".join(sorted(TERM_COLORS.keys()))
        err_msg     = (
            f"Invalid {color_directive.name}: {color_val}. "
            f"Must be 'none', {valid_codes} or a valid color code."
        )
        raise Exception(err_msg)


def _get_default_command(block: parse.Block) -> typ.Optional[str]:
    # TODO (mb 2020-12-17): Allow override/configuration
    lang = block.info_string.strip()
    if lang == 'python':
        # TODO (mb 2020-12-18): sys.executable to work with pypy?
        return "python3"
    elif lang in ("bash", "shell"):
        return "bash"
    elif lang == 'sh':
        return "sh"
    else:
        return None


def _parse_format_options(block: parse.Block) -> common.FormatOptions:
    default_err_fmt = "\u001b[" + TERM_COLORS['red'] + "m{0}\u001b[0m"

    out_fmt = _parse_out_fmt(block, 'out_color', "{0}")
    err_fmt = _parse_out_fmt(block, 'err_color', default_err_fmt)

    out_prefix = get_directive(block, 'out_prefix')
    err_prefix = get_directive(block, 'err_prefix')

    info_directive = get_directive(block, 'proc_info')
    if info_directive is None:
        info_fmt = "# exit: {exit}"
    else:
        info_fmt = info_directive.value

    return common.FormatOptions(
        out=out_fmt,
        err=err_fmt,
        info=info_fmt,
        out_prefix=_parse_prefix(out_prefix) if out_prefix else "",
        err_prefix=_parse_prefix(err_prefix) if err_prefix else "! ",
    )


def _parse_session_block_options(block: parse.Block) -> typ.Optional[common.SessionBlockOptions]:
    exec_directive = get_directive(block, 'exec')
    run_directive  = get_directive(block, 'run')
    out_directive  = get_directive(block, 'out')

    command: typ.Optional[str]

    if exec_directive:
        directive = 'exec'
        command   = exec_directive.value or _get_default_command(block)
    elif run_directive:
        directive = 'run'
        command   = run_directive.value
    elif out_directive:
        directive = 'out'
        command   = None
    else:
        return None

    _provides    = get_directive(block, 'def')
    _debug       = get_directive(block, 'debug')
    _requires    = get_directive(block, 'requires')
    _input_delay = get_directive(block, 'input_delay')
    _timeout     = get_directive(block, 'timeout')
    _expect      = get_directive(block, 'expect')

    provides_id : typ.Optional[str] = None
    if _provides:
        provides_id = _provides.value.strip()
        provides_id = _namespaced_lp_id(block, provides_id)

    if _requires:
        requires_ids = {_namespaced_lp_id(block, req_id) for req_id in _requires.value.split(",")}
    else:
        requires_ids = set()

    timeout_val          = float(_timeout.value    ) if _timeout else DEFUALT_TIMEOUT
    input_delay_val      = float(_input_delay.value) if _input_delay else 0.0
    expected_exit_status = int(_expect.value) if _expect else 0

    return common.SessionBlockOptions(
        command=command,
        directive=directive,
        provides_id=provides_id,
        requires_ids=requires_ids,
        timeout=timeout_val,
        input_delay=input_delay_val,
        expected_exit_status=expected_exit_status,
        is_stdin_writable=directive in ('exec', 'run'),
        is_debug=bool(_debug),
        keepends=True,
        fmt=_parse_format_options(block),
    )


def _parse_capture_output(capture: session.Capture, opts: common.SessionBlockOptions) -> str:
    # TODO: coloring
    # if "\u001b" in capture.stderr:
    #     stderr = capture.stderr
    # elif stderr.strip():
    #     stderr = opts.fmt.err.format(capture.stderr)

    output_lines = []
    for captured_line in capture.lines:
        if captured_line.is_err:
            if opts.fmt.err_prefix:
                line_val = opts.fmt.err_prefix + captured_line.line
            else:
                line_val = captured_line.line
        else:
            if opts.fmt.out_prefix:
                line_val = opts.fmt.out_prefix + captured_line.line
            else:
                line_val = captured_line.line
        output_lines.append(line_val.rstrip())

    output = "\n".join(output_lines)

    if not output.endswith("\n"):
        output += "\n"

    if opts.fmt.info != 'none':
        output += opts.fmt.info.format(
            **{
                'exit'   : capture.exit_status,
                'time'   : capture.runtime,
                'time_ms': capture.runtime * 1000,
            }
        )
        output = output.strip("\r\n")

        # make sure the block closing backticks are
        # on their own line
        if not output.endswith("\n"):
            output += "\n"

    return output


class BlockExecutionError(Exception):
    pass


class BlockTimeoutError(BlockExecutionError):
    pass


TEMPFILE_PATTERN = r"<TEMPFILE([\.\w]+)>"

TEMPFILE_RE = re.compile(TEMPFILE_PATTERN)


def _postproc_failed_block(
    block      : parse.Block,
    opts       : common.SessionBlockOptions,
    stdin_lines: typ.List[str],
    capture    : session.Capture,
) -> None:
    rjust     = len(str(len(stdin_lines)))
    err_lines = []
    for i, line in enumerate(stdin_lines):
        prefix = f"In [{i+1:>{rjust}}]: "
        err_lines.append(prefix + line)
    sys.stderr.write("".join(err_lines) + "\n")
    sys.stderr.flush()

    output       = _parse_capture_output(capture, opts)
    output_lines = output.splitlines()
    rjust        = len(str(len(output_lines)))
    for i, line in enumerate(output_lines):
        prefix = f"Out [{i+1:>{rjust}}]: "
        sys.stderr.write(prefix + line + "\n")
    logger.error(f"Line {block.first_line} of {block.md_path} - Error executing block")


Tempfile = typ.Optional[typ.Any]
Command  = str


def _init_isession(opts: common.SessionBlockOptions, command: Command) -> session.InteractiveSession:
    if opts.is_debug:
        return session.DebugInteractiveSession(command)
    else:
        return session.InteractiveSession(command)


def _init_command(opts: common.SessionBlockOptions) -> typ.Tuple[Tempfile, Command]:
    command = opts.command
    if command is None:
        raise TypeError("Must be str but was None")

    tempfile_placeholder = TEMPFILE_RE.search(command)
    if tempfile_placeholder:
        placeholder = tempfile_placeholder.group(0)
        suffix      = tempfile_placeholder.group(1).split(".", 1)[-1]
        tmp         = tempfile.NamedTemporaryFile(suffix=suffix, delete=False)
        command     = command.replace(placeholder, tmp.name)
        return (tmp, command)
    else:
        return (None, command)


def _process_isession(
    block      : parse.Block,
    opts       : common.SessionBlockOptions,
    command    : str,
    stdin_lines: typ.List[str],
) -> session.Capture:
    _cmd = command if len(command) < 40 else (command[:40] + "...")

    logmsg = f"Line {block.first_line:>5} of {block.md_path} - {opts.directive} {_cmd}"
    logger.info(logmsg)

    isession = _init_isession(opts, command)

    for line in stdin_lines:
        isession.send(line, delay=opts.input_delay)

    try:
        exit_status = isession.wait(timeout=opts.timeout)
    except Exception as ex:
        logger.error(f"Error processing '{command}': {ex}")
        sys.stdout.write("".join(isession.iter_stdout()))
        sys.stderr.write("".join(isession.iter_stderr()))
        raise

    runtime_ms = round(isession.runtime * 1000)

    if exit_status == opts.expected_exit_status:
        exit_info = f"{exit_status}"
    else:
        exit_info = f"{exit_status} != {opts.expected_exit_status}"

    if runtime_ms < 500:
        logger.debug(f"{logmsg:<90} time: {runtime_ms:>6}ms  exit: {exit_info}")
    else:
        logger.info(f"{logmsg:<90} time: {runtime_ms:>6}ms  exit: {exit_info}")

    lines   = isession.output_lines()
    capture = session.Capture(command, exit_status, isession.runtime, lines)

    if exit_status == opts.expected_exit_status:
        # TODO: limit output using max_bytes and max_lines
        # TODO: output escaping/fence style change and errors
        return capture
    else:
        _postproc_failed_block(block, opts, stdin_lines, capture)
        if exit_status == -15:
            raise BlockTimeoutError()
        else:
            raise BlockExecutionError()


def _process_command_block(
    block: parse.Block,
    opts : common.SessionBlockOptions,
) -> session.Capture:
    tmp, command = _init_command(opts)

    if opts.is_stdin_writable:
        stdin_lines = block.inner_content.splitlines(opts.keepends)
    else:
        stdin_lines = []

    if tmp:
        tmp_text = "".join(stdin_lines)
        tmp_data = tmp_text.encode("utf-8")
        tmp.file.write(tmp_data)
        stdin_lines = []

    if opts.directive != 'exec':
        stdin_lines = []

    try:
        return _process_isession(block, opts, command, stdin_lines)
    finally:
        if tmp:
            os.unlink(tmp.name)


def _iter_task_blocks(md_file: parse.MarkdownFile) -> typ.Iterable[common.TaskBlockOpts]:
    for block in md_file.iter_blocks():
        opts = _parse_session_block_options(block)
        if opts:
            yield common.TaskBlockOpts(block, opts)


def _iter_block_tasks(md_file: parse.MarkdownFile) -> typ.Iterable[common.BlockTask]:
    taskblockopt_items = list(_iter_task_blocks(md_file))
    for i, (block, opts) in enumerate(taskblockopt_items):
        capture_index = -1

        if i + 1 < len(taskblockopt_items):
            next_block, next_opts = taskblockopt_items[i + 1]
            if next_opts.directive == 'out':
                capture_index = next_block.elem_index

        command = opts.command
        if command:
            if opts.directive == 'run':
                capture_index = block.elem_index

            yield common.BlockTask(md_file.md_path, command, block, opts, capture_index)


class MarkdownFileResult(typ.NamedTuple):

    orig_md_file    : parse.MarkdownFile
    updated_elements: typ.List[parse.MarkdownElement]


FilesItem = typ.Tuple[parse.MarkdownFile, parse.MarkdownFile]


class BuildOptions(typ.NamedTuple):

    exitfirst      : bool
    in_place_update: bool
    cache_enabled  : bool
    concurrency    : int


class Runner:

    orig_files : MarkdownFiles
    build_files: MarkdownFiles
    opts       : BuildOptions

    results: typ.List[MarkdownFileResult]

    _file_items_by_path: typ.Dict[pl.Path, FilesItem]
    _all_tasks         : typ.List[common.BlockTask]
    _task_results      : typ.List[typ.Tuple[common.BlockTask, session.Capture]]

    _cache: cache.ResultCache

    def __init__(
        self,
        orig_files : MarkdownFiles,
        build_files: MarkdownFiles,
        opts       : BuildOptions,
    ) -> None:
        self.orig_files  = orig_files
        self.build_files = build_files
        self.opts        = opts

        self.results             = []
        self._file_items_by_path = {}
        self._all_tasks          = []
        self._task_results       = []

        # TODO (mb 2021-03-04): _task_results and _cache are somewhat redundant.
        #   It might be possible/reasonable to always rely on the _cache to lookup
        #   the capture of a task, even one that was just executed.

        for orig_md_file, md_file in zip(self.orig_files, self.build_files):
            self._file_items_by_path[md_file.md_path] = (orig_md_file, md_file)
            self._all_tasks.extend(_iter_block_tasks(md_file))

        if self.opts.cache_enabled:
            self._cache = cache.LocalResultCache(self.orig_files)
        else:
            self._cache = cache.DummyCache()

    def _run_task(self, task: common.BlockTask) -> None:
        if self.opts.cache_enabled:
            cached_capture = self._cache.get_capture(task)
        else:
            cached_capture = None

        if cached_capture is None:
            capture = _process_command_block(task.block, task.opts)
            self._cache.update(task, capture)
        else:
            capture = cached_capture

        if task.capture_index >= 0:
            self._task_results.append((task, capture))

    def _postprocess_captures(self) -> None:
        updated_elements = collections.defaultdict(list)

        for task, capture in self._task_results:
            orig_md_file, md_file = self._file_items_by_path[task.md_path]
            output = _parse_capture_output(capture, task.opts)

            elem = md_file.elements[task.capture_index]
            assert elem.md_type == 'block'

            header_lines = [
                line
                for line in elem.content.splitlines(task.opts.keepends)
                if line.startswith("```") or parse.get_line_directive(line, language="shell", is_prelude=True)
            ]

            last_line   = header_lines.pop()
            new_content = "".join(header_lines) + output + last_line

            if elem.content != new_content:
                updated_elem = parse.MarkdownElement(
                    elem.md_path,
                    elem.first_line,
                    elem.elem_index,
                    elem.md_type,
                    new_content,
                    None,
                )
                updated_elements[task.md_path].append(updated_elem)

        for orig_md_file, md_file in zip(self.orig_files, self.build_files):
            result = MarkdownFileResult(orig_md_file, updated_elements[md_file.md_path])
            self.results.append(result)

    def _start(self, submit_task: typ.Callable[[common.BlockTask], Future]) -> None:
        provided_ids        : typ.Set[str] = set()
        remaining_tasks     : typ.List[common.BlockTask] = list(self._all_tasks)
        prev_remaining_tasks: typ.List[common.BlockTask] = []
        num_completed = 0

        while remaining_tasks:
            if remaining_tasks == prev_remaining_tasks:
                for task in remaining_tasks:
                    logger.error(f"Task with unsatisfied requires: {task.opts.requires_ids}")
                raise RuntimeError("No progress made processing block tasks")
            else:
                prev_remaining_tasks = remaining_tasks

            defered_tasks: typ.List[common.BlockTask] = []
            futures      : typ.List[typ.Tuple[Future, common.BlockTask]] = []

            for task in remaining_tasks:
                if task.opts.requires_ids <= provided_ids:
                    future = submit_task(task)
                    futures.append((future, task))
                else:
                    defered_tasks.append(task)

            for future, task in futures:
                future.result(timeout=task.opts.timeout)
                num_completed += 1
                if task.opts.provides_id:
                    provided_ids.add(task.opts.provides_id)

            remaining_tasks = defered_tasks

        logger.info(f"Completed tasks: {num_completed} of {len(self._all_tasks)}")

    def start(self) -> None:
        if self.opts.concurrency == 1 or self.opts.exitfirst:

            def submit(task: common.BlockTask) -> Future:
                future: Future = Future()
                self._run_task(task)
                future.set_result(None)
                return future

            self._start(submit)
        else:
            with ThreadPoolExecutor(max_workers=self.opts.concurrency) as executor:

                def submit(task: common.BlockTask) -> Future:
                    return executor.submit(self._run_task, task)

                self._start(submit)

        self._postprocess_captures()
        self._cache.flush()

    def wait(self) -> None:
        pass


def _run_subprocs(orig_ctx: parse.Context, opts: BuildOptions, runner: Runner) -> parse.Context:
    runner.start()
    runner.wait()

    doc_ctx = orig_ctx.copy()

    for file_idx, md_file_result in enumerate(runner.results):
        if not any(md_file_result.updated_elements):
            continue

        md_path = md_file_result.orig_md_file.md_path

        # phase 6. rewrite output blocks
        orig_elements = list(md_file_result.orig_md_file.elements)
        new_elements  = list(orig_elements)  # copy
        for elem in md_file_result.updated_elements:
            orig_elem = orig_elements[elem.elem_index]
            assert 'out' in orig_elem.content or 'run' in orig_elem.content
            new_elements[elem.elem_index] = elem

        new_md_file = parse.MarkdownFile(md_path, new_elements)
        if opts.in_place_update:
            new_file_content = str(new_md_file)
            with new_md_file.md_path.open(mode="w", encoding="utf-8") as fobj:
                fobj.write(new_file_content)

            logger.info(f"Updated {new_md_file.md_path}")
        else:
            logger.info(f"Update skipped for {new_md_file.md_path} (use -i/--in-place-update)")

        doc_ctx.files[file_idx] = new_md_file

    return doc_ctx


def build(
    orig_ctx: parse.Context,
    opts    : BuildOptions,
) -> parse.Context:
    build_ctx   = orig_ctx.copy()
    build_start = time.time()

    try:
        # NOTE (mb 2020-05-22): macros/constants are abandoned for now
        # phase 1: expand constants
        # build_ctx      = _expand_constants(build_ctx)

        # phase 2: expand dep directives
        expanded_files = list(_iter_expanded_files(build_ctx.files))
    except BlockError as err:
        # TODO (mb 2020-06-03): print context of block
        contents = err.include_contents or [err.block.content]
        for content in contents:
            for line in content.splitlines():
                if line.strip():
                    sys.stderr.write("E " + line.rstrip() + "\n")
        raise

    build_ctx = parse.Context(expanded_files)

    # phase 3. validate blocks
    error_messages = list(_iter_block_errors(orig_ctx, build_ctx))
    for error_msg in error_messages:
        logger.error(error_msg)

    if error_messages:
        raise SystemExit(1)

    # phase 4. write files with expanded blocks
    #   This has to happen before sub-processes, as those
    #   may use the newly created files.
    _dump_files(build_ctx)

    # TODO (mb 2021-01-02): Atomicity guarantees
    #   - Write new files to temp directory
    #   - Only update if original mtimes are still the same

    # phase 5. run sub-processes
    runner = Runner(orig_ctx.files, build_ctx.files, opts)
    try:
        return _run_subprocs(orig_ctx, opts, runner)
    finally:
        duration = time.time() - build_start
        logger.info(f"Build finished after {duration:9.3f}sec")



# This file is part of the litprog project
# https://github.com/litprog/litprog
#
# Copyright (c) 2018-2021 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
import typing as typ
import logging

import markdown as md

log = logging.getLogger(__name__)

MarkdownText = str

HTMLText = str

# https://python-markdown.github.io/extensions/toc/
TocHTML   = str
TocTokens = typ.List[typ.Dict[str, typ.Any]]


class HTMLResult(typ.NamedTuple):

    raw_html  : HTMLText
    toc_html  : TocHTML
    toc_tokens: TocTokens
    filename  : str


def init_md_ctx() -> md.Markdown:
    # https://python-markdown.github.io/extensions/
    extensions = [
        "markdown.extensions.toc",
        "markdown.extensions.extra",
        "markdown.extensions.abbr",
        "markdown.extensions.attr_list",
        "markdown.extensions.def_list",
        "markdown.extensions.fenced_code",
        "markdown.extensions.footnotes",
        "markdown.extensions.tables",
        "markdown.extensions.admonition",
        "markdown.extensions.codehilite",
        "markdown.extensions.meta",
        "markdown.extensions.sane_lists",
        "markdown.extensions.wikilinks",
        "markdown_aafigure",
        "markdown_blockdiag",
        "markdown_svgbob",
        "markdown_katex",
        #####
        # "markdown.extensions.legacy_attr",
        # "markdown.extensions.legacy_em",
        # "markdown.extensions.nl2br",
        # "markdown.extensions.smarty",
    ]

    extension_configs: typ.Dict[str, typ.Any] = {
        'markdown_svgbob': {
            'tag_type'      : "img_base64_svg",
            'bg_color'      : "transparent",
            'fg_color'      : "black",
            'min_char_width': 70,
        },
        'markdown_katex'                : {'no_inline_svg': True, 'insert_fonts_css': False},
        "markdown.extensions.codehilite": {
            # NOTE (mb 2020-06-05): The default guess_lexer=True can detect the
            #   wrong language for certain blocks and colour everything as an
            #   error. The explicit language of the blocks appears to override
            #   this regardless, so this hopefully only affects blocks without
            #   a language in their info string.
            'guess_lang': False,
        },
    }
    return md.Markdown(extensions=extensions, extension_configs=extension_configs)


def md2html(md_text: MarkdownText, filename: str = "<markdown_blob>") -> HTMLResult:
    md_ctx        = init_md_ctx()
    raw_html_text = md_ctx.convert(md_text)
    return HTMLResult(raw_html_text, md_ctx.toc, md_ctx.toc_tokens, filename)  # type: ignore



# This file is part of the litprog project
# https://github.com/litprog/litprog
#
# Copyright (c) 2018-2021 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
import io
import re
import math
import string
import typing as typ
import logging
import itertools as it

import bs4
import pyphen

from . import md2html

logger = logging.getLogger(__name__)


MAX_CODE_BLOCK_LINE_LEN = 85

HTML_PART_PATTERN = re.compile(r"(&[#\w]+?;|<.*?>|\s+\w+)")


HTMLText = str


def _is_entity(part: str) -> int:
    return part.startswith("&") and part.endswith(";")


def _is_tag(part: str) -> int:
    return part.startswith("<") and part.endswith(">")


def _part_len(part: str) -> int:
    if _is_entity(part):
        return 1
    elif _is_tag(part):
        return 0
    else:
        return len(part)


def _iter_new_parts(remaining_part: str, max_len: int) -> typ.Iterable[str]:
    while remaining_part:
        if "://" in remaining_part:
            new_part, remaining_part = remaining_part.split("://", 1)
            new_part += "://"
        elif "?" in remaining_part:
            new_part, remaining_part = remaining_part.split("?", 1)
            new_part += "?"
        elif "/" in remaining_part:
            new_part, remaining_part = remaining_part.split("/", 1)
            new_part += "/"
        elif "#" in remaining_part:
            new_part, remaining_part = remaining_part.split("#", 1)
            new_part += "#"
        elif "." in remaining_part:
            new_part, remaining_part = remaining_part.split(".", 1)
            new_part += "."
        else:
            new_part       = remaining_part[:max_len]
            remaining_part = remaining_part[max_len:]

        # If a line is wrapped, it should never have
        #   trailing whitespace. This gives readers the
        #   best chance of knowing that whitespace exists,
        #   namely by the fact that the wrapped portion of
        #   the line has some indentation
        new_part_rstrip    = new_part.rstrip()
        traling_whitespace = len(new_part) - len(new_part_rstrip)
        if new_part_rstrip and traling_whitespace:
            remaining_part = new_part[-traling_whitespace:] + remaining_part
            new_part       = new_part_rstrip

        yield new_part


def _iter_wrapped_line_parts(line: str, max_len: int) -> typ.Iterable[str]:
    if max_len == 0 or len(line) < max_len:
        yield line
        return

    # Step 1: Split apart whatever we can
    parts = []

    last_end_idx = 0

    for match in HTML_PART_PATTERN.finditer(line):
        begin_idx, end_idx = match.span()
        parts.append(line[last_end_idx:begin_idx])
        parts.append(line[begin_idx   :end_idx  ])
        last_end_idx = end_idx

    parts.append(line[last_end_idx:])

    # Step 2: Split apart whatever we have to
    #   - urls and paths on slash, ? and &
    #   - everything else simply by part[:max_len] part[max_len:]

    for i, part in enumerate(list(parts)):
        if _part_len(part) > max_len:
            split_parts = list(_iter_new_parts(remaining_part=part, max_len=max_len))
            parts[i : i + 1] = split_parts

    if len(parts) == 1:
        yield parts[0]
        return

    chunk: typ.List[str] = []
    chunk_len = 0
    for part in parts:
        if chunk and chunk_len + _part_len(part) >= max_len:
            yield "".join(chunk)
            chunk     = []
            chunk_len = 0

        chunk_len += _part_len(part)
        chunk.append(part)

    if chunk:
        yield "".join(chunk)


BlockLinenos      = typ.List[typ.Tuple[int, int]]
MaybeBlockLinenos = typ.Optional[BlockLinenos]


def iter_wrapped_lines(
    pre_content_text     : str,
    max_line_len         : int  = 80,
    add_line_numbers     : bool = True,
    first_lineno         : int  = 0,
    add_initial_linebreak: bool = False,
) -> typ.Iterable[str]:
    pre_content_text  = pre_content_text.replace("<span></span>", "")
    pre_content_lines = pre_content_text.splitlines()

    # NOTE: code blocks are basically hardcoded to a width of two characters
    # line_number_width = len(str(len(pre_content_lines)))
    for line_idx, line in enumerate(pre_content_lines):
        lineno = first_lineno + line_idx + 1
        parts  = list(_iter_wrapped_line_parts(line, max_len=max_line_len))
        for part_idx, line_part in enumerate(parts):
            if add_line_numbers:
                if part_idx == 0:
                    lineno_span = f"<span class=\"lineno\">{lineno}</span>"
                else:
                    lineno_span = "<span class=\"lineno\">\u21AA</span>"

                if add_initial_linebreak and line_part.startswith("<code"):
                    tag_end_idx = line_part.index(">")
                    # TODO: Fix this cludge. For some reason, the
                    #   first line is indented by less than a full
                    #   space and I have no idea why.
                    _stupid_linebreak = "\n"
                    line_part         = (
                        line_part[: tag_end_idx + 1]
                        + _stupid_linebreak
                        + lineno_span
                        + line_part[tag_end_idx + 1 :]
                    )
                elif line_part != "</code>":
                    yield lineno_span

            if line_part == "</code>":
                # NOTE (mb 2020-06-04): since <code></code> is wrapped with <pre>
                #   we don't want to add extra whitespace at the end.
                yield line_part
            else:
                yield line_part + "\n"


PRE_CODE_BLOCK = """
PYCALVER_REGEX = re.compile(PYCALVER_PATTERN, flags=re.VERBOSE)
INFO    - git tag --annotate v201812.0006-beta --message v201812.0006-beta
INFO    - fetching tags from remote (to turn off use: -n / --no-fetch)
INFO    - git push origin v201812.0006-beta
mypkg  v201812.0665    # last stable release
mypkg  v201812.0666-rc # pre release for testers
mypkg  v201901.0667    # final release after testing period

# bug is discovered in v201812.0666-beta and v201901.0667

mypkg  v201901.0668    # identical code to v201812.0665

# new package is created with compatibility breaking code

mypkg2 v201901.0669    # identical code to v201901.0667
mypkg  v201901.0669    # updated readme, declaring support
                       # level for mypkg, pointing to mypgk2
                       # and documenting how to upgrade.
$ pycalver test 'v201811.0051-beta' '{pycalver}' --release final
"""


# for line_part in iter_wrapped_lines(PRE_CODE_BLOCK.strip(), add_line_numbers=False):
#     print(len(line_part), repr(line_part))
# sys.exit(1)


def _iter_postproc_content_html(
    html_res             : md2html.HTMLResult,
    max_line_len         : int,
    add_initial_linebreak: bool              = False,
    block_linenos        : MaybeBlockLinenos = None,
) -> typ.Iterable[str]:
    content_html: HTMLText = html_res.raw_html
    content_html = content_html.replace("<table>" , """<div class="table-wrap"><table>""")
    content_html = content_html.replace("</table>", """</table></div>""")

    pre_begin_re = re.compile(r'<div class="codehilite"><pre>')
    pre_end_re   = re.compile(r"</pre>")

    last_end_idx = 0
    block_index  = 0

    for match in pre_begin_re.finditer(content_html):
        _begin_lidx, begin_ridx = match.span()
        yield content_html[last_end_idx:begin_ridx]

        end_match = pre_end_re.search(content_html, begin_ridx + 1)
        assert end_match is not None
        end_lidx, end_ridx = end_match.span()
        content_text = content_html[begin_ridx:end_lidx]

        if block_linenos is None:
            first_lineno = 0
        elif block_index < len(block_linenos):
            first_lineno, num_lines = block_linenos[block_index]
            num_content_lines = content_text.strip("\n").count("\n")
            if num_lines == num_content_lines:
                block_index += 1
            else:
                loc = f"{html_res.filename}"
                logger.warning(f"could not match line numbers of block {loc} to html <code>")
                first_lineno = 0
        else:
            loc = f"{html_res.filename}"
            logger.warning(f"could not match line numbers in {loc} to html <code> block")

        wrapped_lines = iter_wrapped_lines(
            content_text,
            max_line_len=max_line_len,
            first_lineno=first_lineno,
            add_initial_linebreak=add_initial_linebreak,
        )
        yield "".join(wrapped_lines)

        end_tag = content_html[end_lidx : end_ridx + 1]
        yield end_tag

        last_end_idx = end_ridx + 1

    yield content_html[last_end_idx:]


SOFT_HYPHEN = "&shy;"


def _shyphenated(dic: pyphen.Pyphen, word: str) -> str:
    word = dic.inserted(word, hyphen=SOFT_HYPHEN)
    if word[2:7] == SOFT_HYPHEN:
        word = word[:2] + word[7:]
    if word[-7:-2] == SOFT_HYPHEN:
        word = word[:-7] + word[-2:]
    return word


WORD_RE = re.compile(r"\w+", flags=re.UNICODE)


def _iter_shyphenated(dic: pyphen.Pyphen, text: str) -> typ.Iterable[str]:
    # NOTE (mb 2021-03-11): While all browsers now (2021 finally!)
    #   have support for "hyphens: auto;", I need to evaluate which
    #   performs better. Superficially the manuall hyphenation here
    #   seemed to work better.
    text = text.replace("\u00AD", "").replace("&shy;", "")

    prev_end = 0
    for match in WORD_RE.finditer(text):
        start, end = match.span()
        if prev_end < start:
            yield text[prev_end:start]

        word = text[start:end]
        if len(word) < 6:
            yield word
        else:
            yield _shyphenated(dic, word)

        prev_end = end

    yield text[prev_end:]


PARSER_MODULE = "html.parser"

INLINE_TAG_NAMES = {"span", "b", "i", "a", "em", "small", "strong", "sub", "sup"}


def _shyphenate(dic: pyphen.Pyphen, text: str) -> str:
    if len(text) < 5:
        return text
    else:
        return "".join(_iter_shyphenated(dic, text))


def _shyphenate_html(soup: bs4.BeautifulSoup) -> None:
    dic = pyphen.Pyphen(lang="en_US")

    elements = it.chain(soup.find_all("p"), soup.find_all("li"))
    for elem in elements:
        for part in elem.contents:
            is_nav_string = isinstance(part, bs4.element.NavigableString)
            is_text_elem  = not is_nav_string and part.name in INLINE_TAG_NAMES and part.string
            if is_nav_string or is_text_elem:
                if is_text_elem:
                    classes  = part.attrs.get('class', ())
                    is_katex = "katex" in classes or "katex-display" in classes
                    if part.name == 'span' and is_katex:
                        continue

                shyphenated = _shyphenate(dic, str(part.string))
                # NOTE: Ugh! So much wrapping just to avoid escaping.
                #   If we don't do this though, we'll get "&shy;" -> "&amp;shy;"
                shy_text = bs4.BeautifulSoup(io.StringIO(shyphenated), PARSER_MODULE)
                part.string.replace_with(shy_text)


def _add_code_scrollers(soup: bs4.BeautifulSoup) -> None:
    for elem in soup.find_all('div', {'class': 'codehilite'}):
        scroller = soup.new_tag('div')
        scroller.attrs['class'] = ['code-scroller']
        list(elem.children)[0].wrap(scroller)


def _update_footnote_refs(soup: bs4.BeautifulSoup) -> None:
    for elem in soup.find_all('a', {'class': 'footnote-ref'}):
        elem.string = "[" + elem.string + "]"


FNOTES_TEXT = "Footnotes and Links"


def _add_footnotes_header(soup: bs4.BeautifulSoup) -> None:
    refs_h        = soup.new_tag('h1')
    refs_h.string = FNOTES_TEXT
    refs_h['id'] = ['references']

    footnotes = soup.find('div', {'class': 'footnote'})
    if footnotes:
        footnotes.insert(2, refs_h)


def _add_footer_links(soup: bs4.BeautifulSoup, fmt: str) -> None:
    hrefs = []
    for a_tag in soup.select('a'):
        href = a_tag.attrs['href']
        if href.startswith("http"):
            hrefs.append(href)

    linklist = soup.new_tag('ol')
    linklist.attrs['class'] = ['linklist']
    footnotes = soup.find('div', {'class': 'footnote'})
    if footnotes:
        footnotes.append(linklist)

    if 'tallcol' in fmt:
        stride = 70
    elif 'letter' in fmt:
        stride = 40
    else:
        stride = 42

    for i, href in enumerate(hrefs):
        linktext        = soup.new_tag('code')
        href_lines      = [href[i : i + stride] for i in range(0, len(href), stride)]
        linktext.string = "\n".join(href_lines)

        link = soup.new_tag('a')
        link['href'] = href
        link.append(linktext)

        li_tag = soup.new_tag('li')
        li_tag.append(link)
        linklist.append(li_tag)


def _add_heading_links(soup: bs4.BeautifulSoup) -> None:
    for heading in soup.select("h1, h2, h3, h4, h5"):
        a_tag = soup.new_tag("a", href="#" + heading['id'])
        if heading.string is None:
            a_tag.extend(list(heading.children))
            heading.clear()
            heading.append(a_tag)
        else:
            heading.string.wrap(a_tag)


def _add_heading_numbers_screen(content_soup: bs4.BeautifulSoup, sections_soup: bs4.BeautifulSoup) -> None:
    numbers_by_hashlink: typ.Dict[str, str] = {
        node['href'].split("#", 1)[-1]: node.text.split(" ", 1)[0] for node in sections_soup.select("a")
    }

    for heading in content_soup.select("h1, h2, h3, h4, h5"):
        number            = numbers_by_hashlink.get(heading['id'], "")
        is_section_number = number and "." in number
        if is_section_number:
            heading.insert(0, number + " ")


def _add_heading_numbers_print(
    soup: bs4.BeautifulSoup, toc_tokens: md2html.TocTokens, heading_prefix: str = ""
) -> None:
    for i, entry in enumerate(toc_tokens):
        heading_number = heading_prefix + str(i + 1)

        tag     = "h" + str(entry['level'])
        heading = soup.find(tag, {'id': entry['id']})
        heading['heading-num'] = heading_number
        if heading_prefix:
            heading.insert(0, heading_number + " ")
        _add_heading_numbers_print(soup, entry['children'], heading_number + ".")


def _add_figure_numbers(soup: bs4.BeautifulSoup) -> None:
    selectors = [
        "h1",
        "p > img",
        ".codehilite",
        ".katex-display",
        "table",
        ".admonition.caption",
    ]
    chapter = 0
    fig_num = -1
    for elem in soup.select(", ".join(selectors)):
        if elem.name == 'h1':
            chapter += 1
            fig_num = -1
        elif 'caption' in elem.get('class', []):
            fig_id      = ""
            cur_fig_num = fig_num
            while cur_fig_num > 0:
                cur_fig_num, rem = divmod(cur_fig_num, 26)
                fig_id = string.ascii_lowercase[rem - 1 if fig_id else rem] + fig_id

            if not fig_id:
                fig_id = "a"

            fig_prefix = "Figure " + str(chapter) + fig_id
            title_elem = elem.find('p', {'class': 'admonition-title'})
            if title_elem:
                title_elem.string = fig_prefix + ": " + (title_elem.string or "")
            else:
                title_elem = soup.new_tag('p')
                title_elem.attrs['class'] = ['admonition-title']
                title_elem.string = fig_prefix
                elem.insert(0, title_elem)
        else:
            fig_num += 1


def _add_nav_numbers(ul_tag: bs4.BeautifulSoup, heading_prefix: str = "", heading: str = "") -> None:
    li_children = [child for child in ul_tag.children if child.name == "li"]

    i = 1
    for li_tag in li_children:
        if heading == li_tag.a.string:
            sub_heading    = heading
            heading_number = sub_heading.split(" ", 1)[0]
        else:
            heading_number  = heading_prefix + str(i)
            sub_heading     = heading_number + " " + li_tag.a.string
            li_tag.a.string = sub_heading
            i += 1

        sub_uls = [child for child in li_tag.children if child.name == "ul"]
        for sub_ul in sub_uls:
            _add_nav_numbers(sub_ul, heading_number + ".", sub_heading)


CODE_CHUNK_SIZE = 5


def _split_code_blocks(soup: bs4.BeautifulSoup) -> None:
    """Split large code blocks so they can span multiple pdf pages."""
    for codehilite in soup.select(".codehilite"):
        code_blocks = codehilite.select("pre > code")
        assert len(code_blocks) == 1
        code_block = code_blocks[0]
        num_lines  = len(code_block.select(".lineno"))
        num_chunks = num_lines / CODE_CHUNK_SIZE
        if num_chunks <= 1:
            continue

        max_chunk_lines = num_lines // math.ceil(num_chunks) + 1
        chunks          = []

        code_nodes = iter(list(code_block.contents))

        try:
            chunk_lines = 0
            chunk       = []
            while True:
                code_node = next(code_nodes)
                is_tag    = isinstance(code_node, bs4.element.Tag)
                if is_tag and 'lineno' in code_node.attrs.get('class'):
                    chunk_lines += 1

                if chunk_lines < max_chunk_lines:
                    chunk.append(code_node)
                else:
                    chunks.append(chunk)

                    chunk_lines = 1
                    chunk       = [code_node]
        except StopIteration:
            if chunk:
                chunks.append(chunk)

        new_pre_nodes = []
        for chunk in chunks:
            new_block          = soup.new_tag('code')
            new_block.contents = chunk
            code_pre           = soup.new_tag('pre')
            code_pre.contents  = [new_block]
            new_pre_nodes.append(code_pre)

        codehilite.contents = new_pre_nodes


def postproc_nav_html(nav_html: HTMLText, has_footnotes: bool = False) -> HTMLText:
    nav_soup = bs4.BeautifulSoup(nav_html, PARSER_MODULE)
    toc_ul   = nav_soup.select(".toc > ul")[0]
    _add_nav_numbers(toc_ul)

    if has_footnotes:
        refs_a        = nav_soup.new_tag('a')
        refs_a.string = FNOTES_TEXT
        refs_a['href'] = "#references"
        refs_li = nav_soup.new_tag('li')
        refs_li.append(refs_a)

        toc_ul.append(refs_li)

    return str(nav_soup)


class HTMLTexts(typ.NamedTuple):
    content : HTMLText
    chapters: HTMLText
    sections: HTMLText

    chapter_prev_href: str
    chapter_next_href: str
    chapter_prev_text: str
    chapter_next_text: str


def postproc4screen(
    html_res     : md2html.HTMLResult,
    block_linenos: BlockLinenos,
    nav_html     : HTMLText,
) -> HTMLTexts:
    # content_html = "".join(_wrap_firstpara(content_html))
    html_chunks = _iter_postproc_content_html(
        html_res,
        max_line_len=MAX_CODE_BLOCK_LINE_LEN,
        add_initial_linebreak=False,
        block_linenos=block_linenos,
    )
    content_html = "".join(html_chunks)

    content_soup = bs4.BeautifulSoup(content_html, PARSER_MODULE)
    nav_soup     = bs4.BeautifulSoup(nav_html    , PARSER_MODULE)

    _add_heading_numbers_screen(content_soup, nav_soup)
    _add_figure_numbers(content_soup)
    _add_heading_links(content_soup)

    # NOTE (mb 2021-04-18): Since chrome finally supports hyphens: auto
    #   on balance is better to disable this. The main reason is that copy
    #   and paste from the html output will include invisible hyphens.
    # _shyphenate_html(content_soup)

    _add_code_scrollers(content_soup)
    _update_footnote_refs(content_soup)
    _add_footnotes_header(content_soup)
    content_html = str(content_soup)

    # content_html = content_html.replace("\u00AD", "&shy;")

    chapters_soup = bs4.BeautifulSoup(str(nav_soup), PARSER_MODULE)
    for elem in chapters_soup.select(".toc a"):
        elem.attrs['href'] = elem.attrs['href'].split("#")[0]

    for elem in chapters_soup.select(".toc > ul > li > ul > li > ul"):
        elem.decompose()

    current_elem = chapters_soup.select(".toc > ul > li > ul > li > a")[0]

    parent       = current_elem.parent.parent.parent
    prev_sibling = parent.previous_sibling
    next_sibling = parent.next_sibling

    while prev_sibling and isinstance(prev_sibling, bs4.element.NavigableString):
        prev_sibling = prev_sibling.previous_sibling
    while next_sibling and isinstance(next_sibling, bs4.element.NavigableString):
        next_sibling = next_sibling.next_sibling

    chapter_prev_href = ""
    chapter_next_href = ""
    chapter_prev_text = ""
    chapter_next_text = ""

    if prev_sibling:
        prev_elem         = prev_sibling.select("a")[0]
        chapter_prev_href = prev_elem.attrs['href'].split("#")[0]
        chapter_prev_text = prev_elem.text
    if next_sibling:
        next_elem         = next_sibling.select("a")[0]
        chapter_next_href = next_elem.attrs['href'].split("#")[0]
        chapter_next_text = next_elem.text

    current_elem.attrs['class'] = "active"
    parent.insert(0, current_elem)

    sections_soup = nav_soup.select(".toc > ul > li > ul")
    if sections_soup:
        sections_html = '<div class="toc">' + str(sections_soup[0]) + "</div>"
    else:
        sections_html = ""

    chapters_html = str(chapters_soup)
    return HTMLTexts(
        content_html,
        chapters_html,
        sections_html,
        chapter_prev_href,
        chapter_next_href,
        chapter_prev_text,
        chapter_next_text,
    )


def postproc4print(html_res: md2html.HTMLResult, fmt: str, block_linenos: BlockLinenos) -> HTMLText:
    # TODO: split code blocks
    # - add ids to headlines
    # - collect links and insert superscript (footnote links)
    if fmt in ("print_a4", "print_letter"):
        max_line_len = MAX_CODE_BLOCK_LINE_LEN
    elif "tallcol" in fmt:
        max_line_len = 72
    elif "ereader" in fmt:
        max_line_len = 72
    else:
        max_line_len = MAX_CODE_BLOCK_LINE_LEN

    html_chunks = _iter_postproc_content_html(
        html_res,
        max_line_len=max_line_len,
        add_initial_linebreak=True,
        block_linenos=block_linenos,
    )
    html_text = "".join(html_chunks)

    content_soup = bs4.BeautifulSoup(html_text, PARSER_MODULE)
    _add_heading_numbers_print(content_soup, html_res.toc_tokens)
    _add_figure_numbers(content_soup)
    _update_footnote_refs(content_soup)
    _add_footnotes_header(content_soup)
    _add_footer_links(content_soup, fmt)
    _split_code_blocks(content_soup)
    html_text = str(content_soup)

    return html_text



# This file is part of the litprog project
# https://github.com/litprog/litprog
#
# Copyright (c) 2018-2021 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT

import typing as typ
import pathlib as pl

from . import parse


class FormatOptions(typ.NamedTuple):

    out : str
    err : str
    info: str

    out_prefix: str
    err_prefix: str


class SessionBlockOptions(typ.NamedTuple):
    """A Session Block based on an 'exec', 'run' or 'out' directive.

    If it is an exec or run directive a session is run and the output is
    captured for later use. If it is an out directive, output from a previous
    capture is used.
    """

    command  : typ.Optional[str]
    directive: str

    # TODO (mb 2021-02-26): These ids are not namespaced (yet)
    #   see litprog.build._namespaced_lp_id and litprog.parse.MarkdownFile.block_namespace()
    provides_id         : typ.Optional[str]
    requires_ids        : typ.Set[str]
    timeout             : float
    input_delay         : float
    expected_exit_status: int

    is_stdin_writable: bool
    is_debug         : bool
    keepends         : bool

    fmt: FormatOptions


class TaskBlockOpts(typ.NamedTuple):
    block: parse.Block
    opts : SessionBlockOptions


ElemIndex = int


class BlockTask(typ.NamedTuple):
    md_path: pl.Path
    command: str
    block  : parse.Block
    opts   : SessionBlockOptions
    # block to which the captured output belongs
    capture_index: ElemIndex



# This file is part of the litprog project
# https://github.com/litprog/litprog
#
# Copyright (c) 2018-2021 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
import os
import re
import uuid
import shlex
import shelve
import typing as typ
import hashlib
import logging
import pathlib as pl
import datetime as dt

from . import parse
from . import common
from . import config
from . import session

# import zlib
# import zstandard


logger = logging.getLogger(__name__)


MarkdownFiles = typ.Iterable[parse.MarkdownFile]

SERIAL_VERSION_ID = 1

CaptureData = bytes


class ManifestEntry(typ.NamedTuple):
    created       : str
    runtime_ms    : int
    capture_size  : int
    capture_digest: str
    task_key      : str
    md_path       : str
    task_info     : str


def init_manifest_entry(
    task: common.BlockTask, task_key: str, capture: session.Capture
) -> typ.Tuple[ManifestEntry, CaptureData]:
    created     = dt.datetime.utcnow().isoformat()
    md_filename = str(task.md_path)
    runtime_ms  = round(capture.runtime * 1000)

    capture_data   = session.dumps_capture(capture)
    capture_size   = len(capture_data)
    capture_digest = hashlib.sha1(capture_data).hexdigest()

    blk       = task.block
    task_info = f"@{blk.first_line:>6}"
    if blk.info_string:
        task_info += " - " + blk.info_string.ljust(9)

    content_lines = blk.inner_content.strip().splitlines()
    if content_lines:
        firstline = content_lines[0]
        task_info += " - " + firstline

    entry = ManifestEntry(
        created,
        runtime_ms,
        capture_size,
        capture_digest,
        task_key,
        md_filename,
        task_info,
    )
    return entry, capture_data


def _manifest_lines(manifest_entries: typ.List[ManifestEntry]) -> typ.Iterable[str]:
    manifest_entries.sort()

    colwidths = [0] * len(ManifestEntry._fields)
    for entry in manifest_entries:
        for i, colwidth in enumerate(colwidths):
            colwidths[i] = max(colwidth, len(str(entry[i])))

    for entry in manifest_entries:
        line = "".join(
            f"{val:<{colwidths[i]}}  " if isinstance(val, str) else f"{val:>{colwidths[i]}}  "
            for i, val in enumerate(entry)
        )
        yield line.rstrip()


def dumps_manifest(manifest_entries: typ.List[ManifestEntry]) -> str:
    if manifest_entries:
        return "\n".join(_manifest_lines(manifest_entries))
    else:
        return ""


def parse_manifest(manifest_text: str) -> typ.Iterable[ManifestEntry]:
    max_split = len(ManifestEntry._fields) - 1
    assert max_split == 6
    for manifest_entry in manifest_text.splitlines():
        (
            created,
            runtime_ms,
            capture_size,
            capture_digest,
            task_key,
            md_path,
            task_info,
        ) = re.split(r"\s+", manifest_entry, max_split)

        yield ManifestEntry(
            created,
            int(runtime_ms),
            int(capture_size),
            capture_digest,
            task_key,
            md_path,
            task_info,
        )


class ResultCache:

    task_keys_by_provide_id: typ.Dict[str, str]

    # used for invalidation (whoever provides must invalidate all requires)
    requires_by_provide_id: typ.Dict[str, typ.List[str]]
    manifest              : typ.List[ManifestEntry]

    def __init__(self, manifest_text: str) -> None:
        self.task_keys_by_provide_id = {}
        self.requires_by_provide_id  = {}

        self.manifest = list(parse_manifest(manifest_text))

        for entry in self.manifest:
            self.task_keys_by_provide_id[entry.task_key] = entry.task_key

    def task_key(self, task: common.BlockTask) -> str:
        requires_digest = hashlib.sha1()

        requires_digest.update(task.block.namespace.encode("utf-8"))
        if task.opts.directive == 'run':
            requires_digest.update(task.command.encode("utf-8"))
            for maybe_path in shlex.split(task.command):
                if os.path.exists(maybe_path):
                    mtime = os.stat(maybe_path).st_mtime
                    requires_digest.update(str(mtime).encode("utf-8"))
        else:
            requires_digest.update(task.block.content.encode("utf-8"))

        for require_id in sorted(task.opts.requires_ids):
            requires_digest.update(require_id.encode("utf-8"))
            # For any require, there MUST have previously have been a
            #   call of ResultCache._reset_task_keys for the
            #   corresponding block with the def/provide. This means
            #   that task_keys_by_provide_id must be populated.
            task_key = self.task_keys_by_provide_id[require_id]
            requires_digest.update(task_key.encode("utf-8"))

        return requires_digest.hexdigest()

    def invalidate_requires(self, provides_id: str) -> None:
        self.task_keys_by_provide_id.pop(provides_id, None)

        for requires_id in self.requires_by_provide_id.get(provides_id, []):
            self.invalidate_requires(requires_id)

    def _reset_task_keys(
        self,
        task : common.BlockTask,
        entry: ManifestEntry,
    ) -> None:
        provides_id = task.opts.provides_id
        if provides_id is None:
            return

        prev_capture_digest = self.task_keys_by_provide_id.get(provides_id)
        if entry.capture_digest != prev_capture_digest:
            self.invalidate_requires(provides_id)
            self.task_keys_by_provide_id[provides_id] = entry.task_key

    def update(self, task: common.BlockTask, capture: session.Capture) -> None:
        task_key = self.task_key(task)
        entry, capture_data = init_manifest_entry(task, task_key, capture)
        self.write_capture(entry, capture_data)
        self._reset_task_keys(task, entry)

    def write_capture(self, entry: ManifestEntry, capture_data: CaptureData) -> None:
        raise NotImplementedError("MUST be implemented by subclass.")

    def read_capture(self, entry: ManifestEntry) -> typ.Optional[CaptureData]:
        raise NotImplementedError("MUST be implemented by subclass.")

    def get_entry(self, task: common.BlockTask) -> typ.Optional[ManifestEntry]:
        task_key = self.task_key(task)
        for entry in reversed(self.manifest):
            if entry.task_key == task_key:
                return entry

        return None

    def get_capture(self, task: common.BlockTask) -> typ.Optional[session.Capture]:
        entry = self.get_entry(task)
        if entry is None:
            return None

        capture_data = self.read_capture(entry)
        if capture_data is None:
            return None

        self._reset_task_keys(task, entry)
        return session.loads_capture(capture_data)

    def flush(self) -> None:
        raise NotImplementedError("MUST be implemented by subclass.")


class DummyCache(ResultCache):
    def __init__(self) -> None:
        super().__init__(manifest_text="")

    def read_capture(self, entry: ManifestEntry) -> typ.Optional[CaptureData]:
        pass

    def write_capture(self, entry: ManifestEntry, capture_data: CaptureData) -> None:
        pass

    def flush(self) -> None:
        pass


# TODO (mb 2021-03-05):

# dctx = zstandard.ZstdDecompressor()
# capture_bytes = dctx.decompress(capture_zstd)

# cctx = zstandard.ZstdCompressor()
# capture_zstd = cctx.compress(capture_bytes)


def _compress(data: bytes) -> bytes:
    return data


def _decompress(data: bytes) -> bytes:
    return data


def parse_cache_id(md_files: MarkdownFiles) -> str:
    """Parse identifier that is unique to the project of the md_files.

    Special consideration is be given to
    derived the cache_id in a machine independent way.

    This leaves open the option of using redis to reuse
    cache results built on other machines.
    """

    prefix          = "unknown_project"
    project_id_data = b"unknown_project"

    for md_file in md_files:
        meta = md_file.parse_front_matter_meta()
        if 'repo_url' in meta:
            url             = meta['repo_url']
            prefix          = url.replace("/", "_").replace(":", "_")
            project_id_data = url.encode("utf-8")
        elif prefix == "unknown_project" and 'title' in meta:
            prefix          = meta['title'].lower().replace(" ", "_")
            project_id_data = meta['title'].encode("utf-8")

    project_id_hash = hashlib.sha1(project_id_data).hexdigest()
    return prefix + "_" + project_id_hash


class LocalResultCache(ResultCache):

    _manifest_file: pl.Path
    _data_file    : pl.Path
    _entry_buffer : typ.List[ManifestEntry]

    def __init__(
        self,
        orig_files: MarkdownFiles,
    ) -> None:
        cache_subdir = parse_cache_id(orig_files)
        cache_dir    = config.CACHE_DIR / cache_subdir

        if not cache_dir.exists():
            cache_dir.mkdir()

        self._manifest_file = cache_dir / f"build_cache.manifest_v{SERIAL_VERSION_ID}"
        self._data_file     = cache_dir / "build_cache.db"

        self._db = shelve.open(str(self._data_file))

        if self._manifest_file.exists():
            with self._manifest_file.open() as fobj:
                manifest_text = fobj.read()
        else:
            manifest_text = ""

        super().__init__(manifest_text)

    def write_capture(self, entry: ManifestEntry, capture_data: CaptureData) -> None:
        self.manifest.append(entry)
        self._db[entry.capture_digest] = _compress(capture_data)

    def read_capture(self, entry: ManifestEntry) -> typ.Optional[CaptureData]:
        raw_capture_data = self._db.get(entry.capture_digest)
        if raw_capture_data:
            return _decompress(raw_capture_data)
        else:
            return None

    def flush(self) -> None:
        self.manifest.sort()
        manifest_text = dumps_manifest(self.manifest)

        # verify round trip
        manifest = list(parse_manifest(manifest_text))
        assert manifest == self.manifest

        tmp_file = str(self._manifest_file) + ".tmp_" + uuid.uuid4().hex
        with open(tmp_file, mode="w") as fobj:
            fobj.write(manifest_text)

        os.rename(tmp_file, self._manifest_file)

        self._db.close()


# class RedisResultCache(ResultCache):
#     pass



# This file is part of the litprog project
# https://gitlab.com/mbarkhau/litprog
#
# Copyright (c) 2019-2020 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT

# pylint: disable=protected-access

import pathlib as pl

import litprog.cli


def test_fs_scanning():
    lit_paths = list(litprog.cli._iter_markdown_filepaths(["lit_v3/"]))

    assert len(lit_paths) > 0
    assert all(isinstance(p, pl.Path) for p in lit_paths)
    assert all(p.suffix == ".md" for p in lit_paths)



# This file is part of the litprog project
# https://github.com/litprog/litprog
#
# Copyright (c) 2018-2021 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT

import typing as typ

Function  = typ.TypeVar('Function', bound=typ.Callable[..., typ.Any])
Decorator = typ.Callable[[Function], Function]


def memo(*, nargs: int, keep_signature: bool, maxsize: int = 0) -> Decorator:
    """Memoization for pure functions to enable code that is fast, lazy and functional.

    Currently functions MUST:
     - have 0 or 1 arguments
     - no kwargs
     - no cache clearing
     - max_size=0 (unbound cache size)
     - function signature is not preserved

    Specialized decorators may be implemented based on knowledge
    about the semantics of the memoized function and its users.

    see adm_common.tuple_memoize
    """

    # NOTE (mb 2021-03-09): We might also inspect nargs
    #   from the function signature. Special care must be
    #   taken when dealing with methods though.
    if nargs > 1:
        raise NotImplementedError("memo(nargs>1)")

    if keep_signature:
        raise NotImplementedError("memo(keep_signature=True)")

    if maxsize > 0:
        raise NotImplementedError("memo(maxsize>0)")

    if maxsize == 0 and nargs == 0:

        def decorator_nargs0_unbounded_cache(func: Function) -> Function:
            cache = []  # type: ignore

            def wrapper():
                if len(cache) == 0:
                    ret = func()
                    cache.append(ret)
                return cache[0]

            return wrapper  # type: ignore

        return decorator_nargs0_unbounded_cache
    elif maxsize == 0 and nargs == 1:

        def decorator_nargs1_unbounded_cache(func: Function) -> Function:
            class UnboundedCache(dict):
                def __missing__(self, key):
                    ret = self[key] = func(key)
                    return ret

            cache = UnboundedCache()
            return cache.__getitem__  # type: ignore

        return decorator_nargs1_unbounded_cache
    else:
        errmsg = f"memo(nargs={nargs}, keep_signature={keep_signature}, maxsize={maxsize})"
        raise NotImplementedError(errmsg)



# This file is part of the litprog project
# https://github.com/litprog/litprog
#
# Copyright (c) 2018-2021 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT

"""Asynchronos polling based file system watching."""
import os
import sys
import glob
import time
import typing as typ
import logging
import pathlib as pl
import threading

logger = logging.getLogger("litprog.watch")


# Can be files, directories or glob strings
PathStrings = typ.Sequence[str]


def unix_ms() -> int:
    return int(round(time.time() * 1000))


def _iter_parent_dirs(path_strs: PathStrings) -> typ.Iterable[pl.Path]:
    for path_str in path_strs:
        path = pl.Path(path_str)
        if path.is_file():
            yield path.parent.absolute()
        elif path.is_dir():
            yield path.absolute()
        else:
            glob_path_strs = glob.glob(path_str)
            if glob_path_strs:
                for glob_path_str in glob_path_strs:
                    glob_path = pl.Path(glob_path_str)
                    if glob_path.is_dir():
                        yield glob_path.absolute()
                    else:
                        yield glob_path.parent.absolute()
            else:
                logger.warning(f"Invalid path: '{path_str}'")


# Debounce will delay invocation of the callback
# until the changes have settled.
DEBOUNCE_DELAY = 500

MIN_SLEEP_DURATION = 20
MAX_SLEEP_DURATION = 500


Change  = typ.Any
Changes = typ.Set[Change]


class Watcher:

    _watch_dirs : typ.List[str]
    _file_mtimes: typ.Dict[str, float]

    def __init__(self, path_strs: PathStrings) -> None:
        self._watch_dirs  = sorted({str(p) for p in _iter_parent_dirs(path_strs)})
        self._file_mtimes = {}
        self.refresh_mtimes()

    def _watch_file(self, path: str, new_changes: typ.Set[Change]) -> None:
        stat      = os.stat(path)
        mtime     = stat.st_mtime
        old_mtime = self._file_mtimes.get(path)
        self._file_mtimes[path] = mtime

        if old_mtime is None:
            new_changes.add(('added', path, stat))
        elif old_mtime != mtime:
            new_changes.add(('modified', path, stat))

    def _walk_dir(self, dirpath: str, new_changes: typ.Set[Change]) -> None:
        for entry in os.scandir(dirpath):
            if entry.is_dir():
                self._walk_dir(entry.path, new_changes)
            else:
                self._watch_file(entry.path, new_changes)

    def refresh_mtimes(self) -> None:
        for watch_dir in self._watch_dirs:
            self._walk_dir(watch_dir, set())

    def watch(self, callback: typ.Callable) -> None:
        def _watch_loop() -> None:
            sleep_duration = MIN_SLEEP_DURATION
            last_change    = unix_ms()
            changes: typ.Set[Change] = set()

            while True:
                new_changes: typ.Set[Change] = set()
                for watch_dir in self._watch_dirs:
                    self._walk_dir(watch_dir, new_changes)

                changes.update(new_changes)

                if changes:
                    now = unix_ms()
                    if now - last_change > DEBOUNCE_DELAY:
                        callback(changes)
                        changes.clear()

                    if new_changes:
                        last_change = now
                    sleep_duration = MIN_SLEEP_DURATION
                else:
                    sleep_duration = min(sleep_duration * 2, MAX_SLEEP_DURATION)

                time.sleep(sleep_duration / 1000)

        try:
            thread        = threading.Thread(target=_watch_loop)
            thread.daemon = True
            thread.start()
            while True:
                time.sleep(1)
        except (KeyboardInterrupt, SystemExit):
            return


def main() -> None:
    watcher = Watcher(sys.argv)
    watcher.watch(print)


if __name__ == '__main__':
    main()



# This file is part of the litprog project
# https://github.com/litprog/litprog
#
# Copyright (c) 2018-2021 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
import os
import sys
import glob
import shutil
import typing as typ
import logging
import pathlib as pl
import tempfile

import click

import litprog.build as lp_build
import litprog.parse as lp_parse
import litprog.watch as lp_watch

try:
    import pretty_traceback

    pretty_traceback.install()
except ImportError:
    pass  # no need to fail because of missing dev dependency


logger = logging.getLogger(__name__)


InputPaths = typ.Sequence[str]

click.disable_unicode_literals_warning = True


class LogConfig(typ.NamedTuple):
    fmt: str
    lvl: int


def _parse_logging_config(verbosity: int) -> LogConfig:
    if verbosity == 0:
        return LogConfig("%(levelname)-7s - %(message)s", logging.WARNING)

    log_format = "%(asctime)s.%(msecs)03d %(levelname)-7s " + "%(name)-16s - %(message)s"
    if verbosity == 1:
        return LogConfig(log_format, logging.INFO)

    assert verbosity >= 2
    return LogConfig(log_format, logging.DEBUG)


_PREV_VERBOSITY: int = -1


def _configure_logging(verbosity: int = 0) -> None:
    # pylint: disable=global-statement
    global _PREV_VERBOSITY

    if verbosity <= _PREV_VERBOSITY:
        # allow function to be called multiple times
        return

    _PREV_VERBOSITY = verbosity

    # remove previous logging handlers
    for handler in list(logging.root.handlers):
        logging.root.removeHandler(handler)

    log_cfg = _parse_logging_config(verbosity)
    logging.basicConfig(level=log_cfg.lvl, format=log_cfg.fmt, datefmt="%Y-%m-%dT%H:%M:%S")


def _iter_markdown_filepaths(input_paths: InputPaths) -> typ.Iterable[pl.Path]:
    for path_str in input_paths:
        path = pl.Path(path_str)
        if path.is_file():
            yield path
        elif path.is_dir():
            for ext in MARKDOWN_FILE_EXTENSIONS:
                for fpath in path.glob(f"**/*.{ext}"):
                    yield fpath
        else:
            glob_path_strs = glob.glob(path_str)
            if glob_path_strs:
                for glob_path_str in glob_path_strs:
                    glob_path = pl.Path(glob_path_str)
                    if glob_path.is_file():
                        yield glob_path
                # TODO (mb 2021-01-01): support globs
            else:
                logger.warning(f"Invalid path: '{path_str}'")


def _get_md_paths(input_paths: InputPaths) -> typ.List[pl.Path]:
    if len(input_paths) == 0:
        click.secho("No markdown files given.", fg='red')
        sys.exit(1)

    md_paths = sorted(_iter_markdown_filepaths(input_paths))
    if len(md_paths) == 0:
        msg = f"No markdown files found for {' '.join(input_paths)}"
        logger.error(msg)
        click.secho(msg, fg='red')
        sys.exit(1)

    return md_paths


# TODO (mb 2021-01-28): These should be parsed from the front matter
SELECTED_FORMATS = [
    'print_a4',
    'print_letter',
    'print_ereader',
    'print_a5',
    'print_booklet_a4',
    'print_halfletter',
    'print_booklet_letter',
    # 'print_twocol_letter',
    # 'print_twocol_a4',
]


def _num_cpus() -> int:
    try:
        # pylint: disable=no-member;    not available on all platforms
        return len(os.sched_getaffinity(0))
    except AttributeError:
        return os.cpu_count() or 1


DEFAULT_CONCURRENCY = max(2, _num_cpus())


def _build(
    input_paths    : InputPaths,
    html           : typ.Optional[str],
    pdf            : typ.Optional[str],
    exitfirst      : bool = False,
    in_place_update: bool = False,
    cache_enabled  : bool = True,
    concurrency    : int  = DEFAULT_CONCURRENCY,
) -> None:
    build_opts = lp_build.BuildOptions(
        exitfirst=exitfirst,
        in_place_update=in_place_update,
        cache_enabled=cache_enabled,
        concurrency=concurrency,
    )

    md_paths = _get_md_paths(input_paths)

    parse_ctx = lp_parse.parse_context(md_paths)
    built_ctx = lp_build.build(parse_ctx, build_opts)

    logger.info("build completed")

    if pdf is None and html is None:
        return

    # NOTE: Since the html is the input for the pdf generation, the
    #   html is generated either way, the only question is if the
    #   output goes to a user specified or to a temporary directory.

    if html is None:
        html            = tempfile.mkdtemp(prefix="litprog_")
        is_html_tmp_dir = True
    else:
        is_html_tmp_dir = False

    html_dir = pl.Path(html)

    # pylint: disable=import-outside-toplevel ; lazy import since we don't always need it
    #   if we were to eagerly import this, then it would slow down every cli invokation,
    #   even those which don't generate --html or --pdf output.
    import litprog.gen_docs as lp_gen_docs

    lp_gen_docs.gen_html(built_ctx, html_dir)

    if pdf:
        pdf_dir = pl.Path(pdf)
        lp_gen_docs.gen_pdf(built_ctx, html_dir, pdf_dir, formats=SELECTED_FORMATS)

    if is_html_tmp_dir:
        shutil.rmtree(html_dir)


_in_path_arg = click.Path(readable=True)
_out_dir_arg = click.Path(file_okay=False, writable=True)

_arg_input_paths = click.argument('input_paths', nargs=-1, type=_in_path_arg)

_opt_html = click.option('--html', nargs=1, type=_out_dir_arg)
_opt_pdf  = click.option('--pdf' , nargs=1, type=_out_dir_arg)

_opt_existfirst = click.option(
    '-e',
    "--exitfirst/--no-exitfirst",
    default=False,
    help="Exit instantly on first error.",
)

_opt_in_place = click.option(
    '-i',
    "--in-place-update",
    is_flag=True,
    default=False,
    help="In place update of lp_out and lp_run blocks in markdown files.",
)

_opt_concurrency = click.option(
    '-n',
    "--concurrency",
    default=DEFAULT_CONCURRENCY,
    help="Number of concurrent processes to execute.",
)

_opt_cache_enabled = click.option(
    "--cache-enabled/--no-cache",
    is_flag=True,
    default=True,
    help="Enable/disable block result cache. Default: enabled",
)

_opt_verbose = click.option('-v', '--verbose', count=True, help="Control log level. -vv for debug level.")


@click.group()
@click.version_option(version="2021.1005-alpha")
@_opt_verbose
def cli(verbose: int = 0) -> None:
    """CLI for litprog."""
    _configure_logging(verbose)


@cli.command()
@_arg_input_paths
@_opt_html
@_opt_pdf
@_opt_existfirst
@_opt_in_place
@_opt_concurrency
@_opt_cache_enabled
@_opt_verbose
def build(
    input_paths    : InputPaths,
    html           : typ.Optional[str],
    pdf            : typ.Optional[str],
    exitfirst      : bool = False,
    in_place_update: bool = False,
    concurrency    : int  = DEFAULT_CONCURRENCY,
    cache_enabled  : bool = True,
    verbose        : int  = 0,
) -> None:
    _configure_logging(verbose)
    _build(input_paths, html, pdf, exitfirst, in_place_update, cache_enabled, concurrency)


@cli.command()
@_arg_input_paths
@_opt_html
@_opt_pdf
@_opt_existfirst
@_opt_in_place
@_opt_concurrency
@_opt_cache_enabled
@_opt_verbose
def watch(
    input_paths    : InputPaths,
    html           : typ.Optional[str],
    pdf            : typ.Optional[str],
    exitfirst      : bool = False,
    in_place_update: bool = False,
    concurrency    : int  = DEFAULT_CONCURRENCY,
    cache_enabled  : bool = True,
    verbose        : int  = 0,
) -> None:
    _configure_logging(verbose)

    if len(input_paths) == 0:
        click.secho("No paths given.", fg='red')
        sys.exit(1)

    # initial build
    try:
        _build(input_paths, html, pdf, exitfirst, in_place_update, cache_enabled, concurrency)
    except lp_build.BlockExecutionError:
        pass

    watcher = lp_watch.Watcher(input_paths)

    def _build_cb(changes) -> None:
        try:
            _build(input_paths, html, pdf, exitfirst, in_place_update, cache_enabled, concurrency)
        except lp_build.BlockExecutionError:
            pass
        # refresh mtimes after build, as they may have changed in the meantime
        watcher.refresh_mtimes()

    watcher.watch(callback=_build_cb)


MARKDOWN_FILE_EXTENSIONS = {
    "markdown",
    "mdown",
    "mkdn",
    "md",
    "mkd",
    "mdwn",
    "mdtxt",
    "mdtext",
    "text",
    "Rmd",
}


if __name__ == '__main__':
    cli()



# This file is part of the litprog project
# https://github.com/litprog/litprog
#
# Copyright (c) 2018-2021 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
#
# Originally based on bumpver/vcs.py of https://github.com/mbarkhau/bumpver
# bumpversion project: https://github.com/peritus/bumpversion
# Copyright (c) 2013-2014 Filip Noetzel - MIT License

"""Minimal Git and Mercirial API.

If terminology for similar concepts differs between git and
mercurial, then the git terms are used. For example "fetch"
(git) instead of "pull" (hg) .
"""

import os
import re
import sys
import json
import shlex
import typing as typ
import logging
import pathlib as pl
import tempfile
import subprocess as sp

logger = logging.getLogger("litprog.vcs")


VCS_SUBCOMMANDS_BY_NAME = {
    'git': {
        'is_usable' : "git rev-parse --git-dir",
        'status'    : "git status --porcelain",
        'head_rev'  : "git rev-parse HEAD",
        'logs'      : "git log --format=fuller --date iso --stat",
        'logs_since': "git log --format=fuller --date iso --stat '{latest_sha}..HEAD'",
    },
    'hg': {
        'is_usable' : "hg root",
        'status'    : "hg status -umard",
        'head_rev'  : "hg --debug id -i",
        'logs'      : "NOT_IMPLEMENTED",
        'logs_since': "NOT_IMPLEMENTED",
    },
}


# commit parsing regexps

COMMIT_RE = re.compile(r"commit ([0-9a-f]*)")
AUTHOR_RE = re.compile(r"Author: (.*) <(.*)>")
DATE_RE   = re.compile(r"AuthorDate: (\d\d\d\d-[0-2][0-9]-[0-3][0-9]) (.*)")

# https://regex101.com/r/Xt8Ua3/1
#
# times_log.png        |     Bin 0 -> 4511 bytes
# stats.txt            |       0
# times.csv            |     170 +
# litprog/__init__.py  |       0
# README.md            |      23 ++++++++---------------
FILE_PAT = r"""
    [ ](.*?)
    [ ]+\|[ ]+
    (\d+|Bin)
    [ ]?
    ([+]+)?
    ([-]+)?
"""
FILE_RE = re.compile(FILE_PAT, re.VERBOSE)


SUMMARY_PAT = r"""
    [ ]
    (\d+[ ]files?[ ]changed)?
    (,[ ]\d+[ ]insertions?\(\+\))?
    (,[ ]\d+[ ]deletions?\(\-\))?
"""
SUMMARY_RE = re.compile(SUMMARY_PAT, re.VERBOSE)


# types

Env = typ.Dict[str, str]


class Change(typ.NamedTuple):
    filename     : str
    changed_lines: int
    added_rel    : int
    deleted_rel  : int


class Commit(typ.NamedTuple):
    commit : str  # sha hex hash
    author : str
    email  : str
    date   : str  # YYYY-MM-DD
    changes: typ.Tuple[Change, ...]


CommitKW = typ.Dict[str, typ.Any]


def dump_commits(commits: typ.Iterable[Commit]) -> typ.Iterable[CommitKW]:
    for commit in commits:
        commit_kw = commit._asdict()
        commit_kw['changes'] = [change._asdict() for change in commit.changes]
        yield commit_kw


def load_commits(commit_kws: typ.Iterable[CommitKW]) -> typ.Iterable[Commit]:
    for commit_kw in commit_kws:
        commit_kw['changes'] = tuple(Change(**change_kw) for change_kw in commit_kw['changes'])
        yield Commit(**commit_kw)


def _is_valid_commit_line(line: str) -> bool:
    return (
        # ignore empty lines
        bool(line.strip())
        # ignore commit message (4 empty spaces)
        and not line.startswith("    ")
        # Merge: xxxx xxxx
        and not line.startswith('Merge')
        and not line.startswith("Commit:")
        and not line.startswith("CommitDate:")
    )


def _parse_change(change_args: typ.Sequence[str]) -> Change:
    if change_args[1] == 'Bin':
        changed_lines = 0
        add_rel       = 0
        del_rel       = 0
    else:
        changed_lines = int(change_args[1])
        add_rel       = len(change_args[2] or "")
        del_rel       = len(change_args[3] or "")

    # TODO (mb 2021-01-21): better parsing of changed filename
    #    '{static => src/litprog/static}/pdf_modal.css',
    return Change(
        filename=change_args[1],
        changed_lines=changed_lines,
        added_rel=add_rel,
        deleted_rel=del_rel,
    )


def parse_commits(lines: typ.Iterable[str]) -> typ.Iterable[Commit]:
    commit_kw: typ.Dict[str, str] = {}
    changes  : typ.List[Change] = []

    for line in lines:
        if _is_valid_commit_line(line):
            match = SUMMARY_RE.match(line)
            if match and any(match.groups()):
                continue

            match = COMMIT_RE.match(line)
            if match:
                if commit_kw:
                    # reached next commit, previous one can be yielded
                    yield Commit(changes=tuple(changes), **commit_kw)

                commit_kw = {'commit': match.group(1)}
                changes   = []
                continue

            match = AUTHOR_RE.match(line)
            # Author: Xxxx Yyyy <xxxx@xxxx.com>
            if match:
                commit_kw['author'] = match.group(1)
                commit_kw['email' ] = match.group(2)
                continue

            match = DATE_RE.match(line)
            if match:
                date = match.group(1)
                commit_kw['date'] = date
                continue

            match = FILE_RE.match(line)
            if match:
                changes.append(_parse_change(match.groups()))
                continue

            logger.warning("Unexpected Line: " + repr(line))

    if commit_kw:
        yield Commit(changes=tuple(changes), **commit_kw)


class VCSAPI:
    """Absraction for git and mercurial."""

    def __init__(self, name: str, subcommands: typ.Dict[str, str] = None):
        self.name = name
        # TODO (mb 2021-01-28): Check if project path should be parameterized.
        self.project_path = pl.Path(os.getcwd()).absolute()

        if subcommands is None:
            self.subcommands = VCS_SUBCOMMANDS_BY_NAME[name]
        else:
            self.subcommands = subcommands

    def __call__(self, cmd_name: str, env: Env = None, **kwargs: str) -> str:
        """Invoke subcommand and return output."""
        prev_cwd = pl.Path(os.getcwd()).absolute()
        try:
            if prev_cwd != self.project_path:
                os.chdir(str(self.project_path))

            cmd_tmpl  = self.subcommands[cmd_name]
            cmd_str   = cmd_tmpl.format(**kwargs)
            cmd_parts = shlex.split(cmd_str)
            output_data: bytes = sp.check_output(cmd_parts, env=env, stderr=sp.STDOUT)
            return output_data.decode("utf-8")
        finally:
            if prev_cwd != self.project_path:
                os.chdir(str(prev_cwd))

    def is_usable(self) -> bool:
        """Detect availability of subcommand."""
        if not os.path.exists(f".{self.name}"):
            return False

        cmd = self.subcommands['is_usable'].split()

        try:
            retcode = sp.call(cmd, stderr=sp.PIPE, stdout=sp.PIPE)
            return retcode == 0
        except OSError as err:
            if err.errno == 2:
                # git/mercurial is not installed.
                return False
            else:
                raise

    def status(self) -> typ.List[str]:
        """Get status lines."""
        status_output = self('status')
        status_items  = [line.split(" ", 1) for line in status_output.splitlines()]
        return [filepath.strip().split(" ", 1)[-1] for status, filepath in status_items if status != "??"]

    def head_rev(self) -> str:
        output = self('head_rev')
        return output.strip()

    def _tmp_filepath(self) -> pl.Path:
        cache_dir = "litprog_tmp_20210114t2006"  # change this if the cache format changes
        cache_id, _ = re.subn(r"[^\w]", "_", str(self.project_path))
        cache_filename = "commits_cache_" + cache_id + ".json"
        cache_dirpath  = pl.Path(tempfile.gettempdir()) / cache_dir
        if not cache_dirpath.exists():
            cache_dirpath.mkdir()
        return cache_dirpath / cache_filename

    def _write_commits(self, commits: typ.List[Commit]) -> None:
        cache_data = json.dumps(list(dump_commits(commits)))
        with self._tmp_filepath().open(mode="w") as fobj:
            fobj.write(cache_data)

    def commits(self) -> typ.List[Commit]:
        if self._tmp_filepath().exists():
            with self._tmp_filepath().open(mode="r") as fobj:
                commits = list(load_commits(json.load(fobj)))
            output      = self('logs_since', latest_sha=commits[0].commit)
            log_lines   = output.splitlines()
            new_commits = list(parse_commits(log_lines))
            if new_commits:
                commits = new_commits + commits
                self._write_commits(commits)
            return commits
        else:
            output    = self('logs')
            log_lines = output.splitlines()
            commits   = list(parse_commits(log_lines))
            self._write_commits(commits)
            return commits

    def __repr__(self) -> str:
        """Generate string representation."""
        return f"VCSAPI(name='{self.name}')"


def get_vcs_api() -> typ.Optional[VCSAPI]:
    """Detect the appropriate VCS for a repository.

    raises OSError if the directory doesn't use a supported VCS.
    """
    for vcs_name in VCS_SUBCOMMANDS_BY_NAME:
        vcs_api = VCSAPI(name=vcs_name)
        if vcs_api.is_usable():
            return vcs_api

    return None


def main() -> int:
    vcs_api = get_vcs_api()
    if vcs_api is None:
        print("no vcs")
    else:
        print("HEAD:", vcs_api.head_rev())
        for fpath in vcs_api.status():
            print("    ", repr(fpath))
    return 0


if __name__ == '__main__':
    import pretty_traceback

    pretty_traceback.install(envvar='ENABLE_PRETTY_TRACEBACK')
    sys.exit(main())



#!/home/mbarkhau/miniconda3/envs/litprog_py38/bin/python
# This file is part of the litprog project
# https://github.com/litprog/litprog
#
# Copyright (c) 2018-2021 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
import sys
import math
import time
import typing as typ
import logging
import pathlib as pl

import PyPDF2 as pypdf

logger = logging.getLogger("pdf_booklet")


MAX_BOOKLET_SHEETS = 13
PAGES_PER_SHEET    = 4

PT_PER_INCH = 72
PT_PER_MM   = PT_PER_INCH / 25.4


BOOKLET_FORMAT_MAPPING = {
    # out_format, page_order, scale, margin
    'A6-Portrait'    : ('A5-Landscape'    , 'booklet', 0 * PT_PER_MM),
    'A5-Portrait'    : ('A4-Landscape'    , 'booklet', 0 * PT_PER_MM),
    'A4-Portrait'    : ('A4-Landscape'    , 'booklet', 0 * PT_PER_MM),
    'A3-Portrait'    : ('A4-Landscape'    , 'booklet', 0 * PT_PER_MM),
    'Half-Letter'    : ('Letter-Landscape', 'booklet', 0 * PT_PER_MM),
    '1of2col-A4'     : ('A4-Portrait'     , 'paper'  , 0 * PT_PER_MM),
    '1of2col-Letter' : ('Letter-Portrait' , 'paper'  , 0 * PT_PER_MM),
    'Letter-Portrait': ('A4-Landscape'    , 'booklet', -5 * PT_PER_MM),
}


PAPER_FORMATS_PT = {
    'A6-Portrait'     : (105  * PT_PER_MM  , 148 * PT_PER_MM),
    'A5-Landscape'    : (210  * PT_PER_MM  , 148 * PT_PER_MM),
    'A5-Portrait'     : (148  * PT_PER_MM  , 210 * PT_PER_MM),
    'A4-Landscape'    : (297  * PT_PER_MM  , 210 * PT_PER_MM),
    'A4-Portrait'     : (210  * PT_PER_MM  , 297 * PT_PER_MM),
    'A3-Portrait'     : (297  * PT_PER_MM  , 420 * PT_PER_MM),
    '1of2col-A4'      : (105  * PT_PER_MM  , 297 * PT_PER_MM),
    '1of2col-Letter'  : (4.25 * PT_PER_INCH, 11  * PT_PER_INCH),
    'Half-Letter'     : (5.5  * PT_PER_INCH, 8.5 * PT_PER_INCH),
    'Letter-Landscape': (11   * PT_PER_INCH, 8.5 * PT_PER_INCH),
    'Letter-Portrait' : (8.5  * PT_PER_INCH, 11  * PT_PER_INCH),
}


# https://helpx.adobe.com/acrobat/kb/print-booklets-acrobat-reader.html
# https://tex.stackexchange.com/questions/60934/


def calc_booklet_page_counts(total_pages: int, max_sheets: int = MAX_BOOKLET_SHEETS) -> typ.List[int]:
    total_sheets = math.ceil(total_pages  / PAGES_PER_SHEET)
    num_booklets = math.ceil(total_sheets / max_sheets)

    # spb: sheets per booklet
    # ppb: pages per booklet

    target_ppb = total_pages / num_booklets
    spb        = math.floor(target_ppb / PAGES_PER_SHEET)
    ppb        = spb * PAGES_PER_SHEET

    if ppb * num_booklets < total_pages - PAGES_PER_SHEET:
        ppb += PAGES_PER_SHEET

    if ppb * num_booklets < total_pages:
        # Add extra pages to the last booklet
        last_ppb = ppb + PAGES_PER_SHEET
    else:
        last_ppb = ppb

    return [ppb] * (num_booklets - 1) + [last_ppb]


# Glossary
# Booklet: Multiple sheets that get bound together
# Sheet: Duplex printed piece of paper, with four pages
# Half-Sheet: One face of a sheet with two pages
# Page: One half of one side of a sheet

_BOOKLET_PAGE_NUMBERING_ILLUSTRATION = r"""

         Booklet 1        Booklet 2

         +       +        +       +
    8   /|----->/|   16  /|----->/|
    -->/ | 6   / |   -->/ | 14  / |
      / 7| -->/ 5|     /15| -->/13|
     +   |   +   |    +   |   +   |
     |\  |   |\  |    |\  |   |\  |
     | \/    | \/     | \/    | \/
     |  \    |  \     |  \    |  \
     |   |   |   |    |   |   |   |
     |  1|   |  3|    |  9|   | 11|
      \  |  2 \  |  4  \  | 10 \  | 12
       \ |<--  \ |<--   \ |<--  \ |<--
        \|----->\|       \|----->\|
         +       +        +       +
"""


def booklet_page_layout(
    total_pages: int, max_sheets: int = MAX_BOOKLET_SHEETS
) -> typ.Tuple[typ.List[int], typ.List[int]]:
    booklet_page_counts = calc_booklet_page_counts(total_pages, max_sheets)

    booklet_index_by_page     : typ.List[int] = []
    booklet_page_index_by_page: typ.List[int] = []

    doc_page_offset = 0
    for booklet_index, booklet_page_count in enumerate(booklet_page_counts):
        left_doc_page_index  = doc_page_offset + booklet_page_count - 1
        right_doc_page_index = doc_page_offset
        while right_doc_page_index < left_doc_page_index:
            booklet_page_index_by_page.append(left_doc_page_index)
            booklet_page_index_by_page.append(right_doc_page_index)
            booklet_page_index_by_page.append(right_doc_page_index + 1)
            booklet_page_index_by_page.append(left_doc_page_index - 1)
            booklet_index_by_page += [booklet_index] * 4
            left_doc_page_index -= 2
            right_doc_page_index += 2

        doc_page_offset += booklet_page_count

    assert len(booklet_page_index_by_page) == len(booklet_index_by_page)
    assert len(booklet_page_index_by_page) == len(set(booklet_page_index_by_page))

    return booklet_page_index_by_page, booklet_index_by_page


def get_format_id(
    page_width_pt : float,
    page_height_pt: float,
    epsilon_pt    : float = 2.0,
) -> typ.Optional[str]:
    best_format_id: typ.Optional[str] = None
    best_delta = 999999

    for format_id, (fmt_width_pt, fmt_height_pt) in PAPER_FORMATS_PT.items():
        delta = abs(fmt_width_pt - page_width_pt) + abs(fmt_height_pt - page_height_pt)
        if delta < best_delta:
            best_format_id = format_id

        if delta < epsilon_pt:
            return format_id

    print(best_format_id, delta)
    return None


PDFReader = typ.Any

PDFPage = typ.Any

MediaBox = typ.Any

PageIndexMapping = typ.List[typ.Tuple[int, int]]


class OutputParameters(typ.NamedTuple):

    page_order: str
    width     : float
    height    : float

    scale     : float
    pad_x     : float
    pad_y     : float
    pad_center: float


def _init_output_parameters(reader: PDFReader, out_sheet_format: str) -> OutputParameters:
    # NOTE (mb 2021-03-04): Can we figure out the bounds of the inner content
    #   and scale + translate each page to be in a well defined content box
    #   in the output page?
    # NOTE (mb 2021-03-19): Nope, we can't, at least not easilly with PyPDF2.
    #   A possible alternative is PyMuPDF: https://pymupdf.readthedocs.io/

    # pylint: disable=too-many-locals

    media_box = reader.getPage(0).mediaBox

    in_page_width  = float(media_box.getWidth())
    in_page_height = float(media_box.getHeight())

    out_sheet_width, out_sheet_height = PAPER_FORMATS_PT[out_sheet_format]
    if out_sheet_width < out_sheet_height:
        errmsg = f"Invalid out_sheet_format={out_sheet_format}. Landscape format required."
        raise ValueError(errmsg)

    out_page_width  = out_sheet_width / 2
    out_page_height = out_sheet_height

    x_scale = out_page_width  / in_page_width
    y_scale = out_page_height / in_page_height

    scale = min(x_scale, y_scale)

    scaled_page_width  = in_page_width  * scale
    scaled_page_height = in_page_height * scale

    x_padding = out_page_width  - scaled_page_width
    y_padding = out_page_height - scaled_page_height

    pad_center = x_padding / 4
    pad_center = 0
    pad_x      = (x_padding - pad_center) / 2
    pad_y      = y_padding / 2

    _iw_mm = round(in_page_width      / PT_PER_MM)
    _ih_mm = round(in_page_height     / PT_PER_MM)
    _sw_mm = round(scaled_page_width  / PT_PER_MM)
    _sh_mm = round(scaled_page_height / PT_PER_MM)
    _ow_mm = round(out_sheet_width    / PT_PER_MM)
    _oh_mm = round(out_sheet_height   / PT_PER_MM)
    logger.info("OutputParameters")
    logger.info(f"    scale: {scale:5.2f}x")
    logger.info(f"    in : {_iw_mm}mm x {_ih_mm}mm -> {_sw_mm}mm x {_sh_mm}mm (2x)")
    logger.info(f"    out: {_ow_mm}mm x {_oh_mm}mm")

    return OutputParameters(
        'booklet',
        out_sheet_width,
        out_sheet_height,
        scale,
        pad_x,
        pad_y,
        pad_center,
    )


def _create_sheets(
    in_pages            : typ.List[PDFPage],
    output              : pypdf.PdfFileWriter,
    out_coords          : OutputParameters,
    half_page_to_in_page: PageIndexMapping,
) -> typ.List[PDFPage]:
    sheet_indexes = {half_page_index // 2 for in_page_index, half_page_index in half_page_to_in_page}
    out_sheets    = [
        output.addBlankPage(width=out_coords.width, height=out_coords.height) for _ in sheet_indexes
    ]

    for half_page_index, in_page_index in half_page_to_in_page:
        if len(in_pages) - 1 < in_page_index:
            continue

        in_page = in_pages[in_page_index]

        out_sheet = out_sheets[half_page_index // 2]
        if half_page_index % 2 == 0:
            x_offset = 0 - out_coords.pad_center
        else:
            x_offset = (out_coords.width / 2) + out_coords.pad_center

        translate_x = x_offset + out_coords.pad_x
        translate_y = 0        + out_coords.pad_y

        tzero = time.time()

        if abs(out_coords.scale - 1) < 0.01:
            out_sheet.mergeTranslatedPage(in_page, tx=translate_x, ty=translate_y, expand=False)
        else:
            out_sheet.mergeScaledTranslatedPage(
                in_page, scale=out_coords.scale, tx=translate_x, ty=translate_y, expand=False
            )

        duration = int((time.time() - tzero) * 1000)

        sheet_index = half_page_index % 2
        logger.debug(f"booklet page: {half_page_index:>2} sheet: {sheet_index:>2} {duration:>5}ms")

    return out_sheets


def create(
    in_path         : pl.Path,
    out_path        : typ.Optional[pl.Path] = None,
    out_sheet_format: str = 'A4-Landscape',
) -> pl.Path:
    assert out_sheet_format in PAPER_FORMATS_PT

    max_sheets = MAX_BOOKLET_SHEETS

    if out_path is None:
        ext          = in_path.suffixes[-1]
        out_filename = (in_path.name[: -len(ext)] + "_booklet") + ext
        _out_path    = in_path.parent / out_filename
    else:
        _out_path = out_path

    output = pypdf.PdfFileWriter()

    with in_path.open(mode="rb") as in_fobj:
        reader = pypdf.PdfFileReader(in_fobj)

        output_params = _init_output_parameters(reader, out_sheet_format)

        in_pages = list(reader.pages)

        if output_params.page_order == 'booklet':
            layout = booklet_page_layout(len(in_pages), max_sheets=max_sheets)
            booklet_page_indexes, _booklet_index_by_page = layout
            half_page_to_in_page = list(enumerate(booklet_page_indexes))
        else:
            half_page_to_in_page = [
                (half_page_index, half_page_index) for half_page_index in range(len(in_pages))
            ]

        out_sheets = _create_sheets(in_pages, output, output_params, half_page_to_in_page)
        tzero      = time.time()
        for out_sheet in out_sheets:
            out_sheet.compressContentStreams()
        compression_time = time.time() - tzero
        logger.debug(f"compression time: {compression_time}")

        with _out_path.open(mode="wb") as out_fobj:
            output.write(out_fobj)

    logger.info(f"Wrote to '{_out_path}'")

    return _out_path


def main() -> int:
    logging.basicConfig()
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.INFO)

    create(in_path=pl.Path(sys.argv[1]))
    return 0


if __name__ == '__main__':
    sys.exit(main())



# This file is part of the litprog project
# https://github.com/litprog/litprog
#
# Copyright (c) 2018-2021 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
import os
import pathlib as pl

HOME = os.path.expanduser("~")

FALLBACK_CONFIG_HOME = os.path.join(HOME, ".config")
FALLBACK_CACHE_HOME  = os.path.join(HOME, ".cache")
FALLBACK_DATA_HOME   = os.path.join(HOME, ".local", "share")

CONFIG_HOME = pl.Path(os.getenv('XDG_CONFIG_HOME', FALLBACK_CONFIG_HOME))
CACHE_HOME  = pl.Path(os.getenv('XDG_CACHE_HOME' , FALLBACK_CACHE_HOME ))
DATA_HOME   = pl.Path(os.getenv('XDG_DATA_HOME'  , FALLBACK_DATA_HOME  ))

CONFIG_DIR  = CONFIG_HOME / "litprog"
CACHE_DIR   = CACHE_HOME  / "litprog"
ENVDIR_BASE = DATA_HOME   / "litprog"



# This file is part of the litprog project
# https://gitlab.com/mbarkhau/litprog
#
# Copyright (c) 2019-2020 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT

# pylint: disable=protected-access

import io

import litprog.session as sut

RAW_TEST_TEXT = b"""
Hello \xe4\xb8\x96\xe7\x95\x8c!
foo bar
"""


def test_gen_captured_lines():
    raw_lines      = RAW_TEST_TEXT.strip().splitlines()
    captured_lines = sut._gen_captured_lines(raw_lines)
    lines          = [cl.line for cl in captured_lines]
    assert lines == ["Hello 世界!", "foo bar"]


def test_start_reader():
    expected = RAW_TEST_TEXT.decode("utf-8")

    raw_text_buf   = io.BytesIO(RAW_TEST_TEXT)
    capture_thread = sut._start_reader(raw_text_buf)
    capture_thread.thread.join()

    content = "".join(cl.line for cl in capture_thread.lines)
    assert content == expected


BLOCK_0 = r"""
import sys

def out(s: str):
    sys.stdout.write(s)
    sys.stdout.flush()

def err(s: str):
    sys.stderr.write(s)
    sys.stderr.flush()
"""

BLOCK_1 = r"""
out("ok1\n")
err("moep\n")
assert True
out("ok2")
"""

BLOCK_2 = r"""
out("ok3")
sys.exit(0)
import time
time.sleep(2)
assert False
out("ok4")
"""


def test_integration():
    session = sut.InteractiveSession(cmd=['python3'])
    for block in [BLOCK_0, BLOCK_1, BLOCK_2]:
        session.send(block)
    retcode = session.wait()
    assert session.stdout == "ok1\nok2ok3"
    assert session.stderr == "moep\n"
    assert retcode        == 0
    assert session.runtime < 0.5
